from dataclasses import dataclass
from typing import override

from ai.backend.common.data.permission.types import EntityType
from ai.backend.common.types import AccessKey, KernelId
from ai.backend.manager.actions.action import BaseActionResult
from ai.backend.manager.actions.types import ActionOperationType
from ai.backend.manager.data.session.types import SessionData
from ai.backend.manager.services.session.base import SessionAction


@dataclass
class GetContainerLogsAction(SessionAction):
    session_name: str
    owner_access_key: AccessKey
    kernel_id: KernelId | None

    @override
    @classmethod
    def entity_type(cls) -> EntityType:
        return EntityType.SESSION_CONTAINER_LOG

    @override
    def entity_id(self) -> str | None:
        return None

    @override
    @classmethod
    def operation_type(cls) -> ActionOperationType:
        return ActionOperationType.GET


@dataclass
class GetContainerLogsActionResult(BaseActionResult):
    result: dict[str, dict[str, str]]
    session_data: SessionData

    @override
    def entity_id(self) -> str | None:
        return str(self.session_data.id)



from __future__ import annotations

from pydantic import BaseModel, ConfigDict

__all__ = ("NumberFormat",)


class NumberFormat(BaseModel):
    """Display number format configuration for a resource slot type."""

    model_config = ConfigDict(frozen=True)

    binary: bool = False
    round_length: int = 0



from __future__ import annotations

from dataclasses import dataclass
from typing import override

from ai.backend.manager.actions.action import BaseActionResult
from ai.backend.manager.actions.types import ActionOperationType
from ai.backend.manager.models.scaling_group import ScalingGroupForDomainRow
from ai.backend.manager.repositories.base.purger import BatchPurger

from .domain_base import ScalingGroupDomainAction


@dataclass
class DisassociateScalingGroupWithDomainsAction(ScalingGroupDomainAction):
    """Action to disassociate a scaling group from multiple domains."""

    purger: BatchPurger[ScalingGroupForDomainRow]

    @override
    @classmethod
    def operation_type(cls) -> ActionOperationType:
        return ActionOperationType.DELETE

    @override
    def entity_id(self) -> str | None:
        return None


@dataclass
class DisassociateScalingGroupWithDomainsActionResult(BaseActionResult):
    """Result of disassociating a scaling group from domains."""

    @override
    def entity_id(self) -> str | None:
        return None



import uuid
from dataclasses import dataclass
from typing import override

from ai.backend.manager.actions.action import BaseActionResult
from ai.backend.manager.actions.types import ActionOperationType
from ai.backend.manager.data.artifact.types import ArtifactData
from ai.backend.manager.services.artifact.actions.base import ArtifactAction


@dataclass
class DeleteArtifactsAction(ArtifactAction):
    artifact_ids: list[uuid.UUID]

    @override
    def entity_id(self) -> str | None:
        return None

    @override
    @classmethod
    def operation_type(cls) -> ActionOperationType:
        return ActionOperationType.DELETE


@dataclass
class DeleteArtifactsActionResult(BaseActionResult):
    artifacts: list[ArtifactData]

    @override
    def entity_id(self) -> str | None:
        return None



from __future__ import annotations

from collections.abc import AsyncIterator
from contextlib import asynccontextmanager

from ai.backend.common.etcd import AsyncEtcd
from ai.backend.common.health_checker import ServiceHealthChecker
from ai.backend.common.health_checker.checkers.etcd import EtcdHealthChecker
from ai.backend.manager.config.bootstrap import BootstrapConfig

from .base import BootstrapDependency


class EtcdDependency(BootstrapDependency[AsyncEtcd]):
    """Provides etcd client lifecycle management for bootstrap stage."""

    @property
    def stage_name(self) -> str:
        return "etcd"

    @asynccontextmanager
    async def provide(self, setup_input: BootstrapConfig) -> AsyncIterator[AsyncEtcd]:
        """Initialize and provide an etcd client.

        Args:
            setup_input: Bootstrap configuration containing etcd settings

        Yields:
            Initialized AsyncEtcd client
        """
        async with AsyncEtcd.create_from_config(setup_input.etcd.to_dataclass()) as etcd:
            yield etcd

    def gen_health_checkers(self, resource: AsyncEtcd) -> ServiceHealthChecker:
        """
        Return health checker for etcd.

        Args:
            resource: The initialized etcd client

        Returns:
            EtcdHealthChecker for etcd
        """
        return EtcdHealthChecker(etcd=resource)



from typing import override

from ai.backend.common.contexts.user import current_user
from ai.backend.manager.actions.action import BaseActionTriggerMeta
from ai.backend.manager.actions.action.scope import BaseScopeAction
from ai.backend.manager.actions.validator.scope import ScopeActionValidator
from ai.backend.manager.data.permission.id import ScopeId
from ai.backend.manager.data.permission.role import ScopePermissionCheckInput
from ai.backend.manager.errors.user import UserNotFound
from ai.backend.manager.repositories.permission_controller.repository import (
    PermissionControllerRepository,
)


class ScopeActionRBACValidator(ScopeActionValidator):
    def __init__(
        self,
        repository: PermissionControllerRepository,
    ) -> None:
        self._repository = repository

    @override
    async def validate(self, action: BaseScopeAction, meta: BaseActionTriggerMeta) -> None:
        entity_type = action.entity_type()
        scope_type = action.scope_type()
        scope_id = action.scope_id()
        user = current_user()
        if user is None:
            raise UserNotFound("User not found in context")

        await self._repository.check_permission_in_scope(
            ScopePermissionCheckInput(
                user_id=user.user_id,
                operation=action.operation_type().to_permission_operation(),
                target_entity_type=entity_type,
                target_scope_id=ScopeId(
                    scope_type=scope_type,
                    scope_id=scope_id,
                ),
            )
        )



"""Handler for sweeping stale sessions."""

from __future__ import annotations

import logging
from collections.abc import Sequence

from ai.backend.logging import BraceStyleAdapter
from ai.backend.manager.data.kernel.types import KernelStatus
from ai.backend.manager.data.session.types import SessionStatus, StatusTransitions, TransitionStatus
from ai.backend.manager.defs import LockID
from ai.backend.manager.repositories.scheduler.repository import SchedulerRepository
from ai.backend.manager.sokovan.data import SessionWithKernels
from ai.backend.manager.sokovan.scheduler.handlers.base import SessionLifecycleHandler
from ai.backend.manager.sokovan.scheduler.results import (
    SessionExecutionResult,
    SessionTransitionInfo,
)

log = BraceStyleAdapter(logging.getLogger(__name__))


class SweepSessionsLifecycleHandler(SessionLifecycleHandler):
    """Handler for sweeping stale sessions (maintenance operation).

    Following the DeploymentCoordinator pattern:
    - Coordinator queries sessions with PENDING status (provides HandlerSessionData)
    - Handler fetches detailed timeout data and determines which sessions have timed out
    - Stale sessions are moved to TERMINATING status
    """

    def __init__(
        self,
        repository: SchedulerRepository,
    ) -> None:
        self._repository = repository

    @classmethod
    def name(cls) -> str:
        """Get the name of the handler."""
        return "sweep-sessions"

    @classmethod
    def target_statuses(cls) -> list[SessionStatus]:
        """Sessions that may need sweeping - PENDING with timeout."""
        return [SessionStatus.PENDING]

    @classmethod
    def target_kernel_statuses(cls) -> list[KernelStatus] | None:
        """No kernel filtering for sweep check."""
        return None

    @classmethod
    def status_transitions(cls) -> StatusTransitions:
        """Define state transitions for sweep sessions handler (BEP-1030).

        - success: None (sweep operation doesn't have success transition)
        - need_retry: None
        - expired: Session/kernel → TERMINATING (pending timeout exceeded)
        - give_up: None
        """
        return StatusTransitions(
            success=None,
            need_retry=None,
            expired=TransitionStatus(
                session=SessionStatus.TERMINATING,
                kernel=KernelStatus.TERMINATING,
            ),
            give_up=None,
        )

    @property
    def lock_id(self) -> LockID | None:
        """No lock needed for sweeping stale sessions."""
        return None

    async def execute(
        self,
        _scaling_group: str,
        sessions: Sequence[SessionWithKernels],
    ) -> SessionExecutionResult:
        """Sweep stale sessions including those with pending timeout.

        The coordinator provides SessionWithKernels with full SessionInfo/KernelInfo.
        This handler:
        1. Fetches detailed timeout data from repository
        2. Adds timed out sessions to stales for Coordinator to handle status transition
        """
        result = SessionExecutionResult()

        if not sessions:
            return result

        # Extract session IDs from SessionWithKernels
        session_ids = [s.session_info.identity.id for s in sessions]

        # Fetch detailed session data with timeout info
        timed_out_sessions = await self._repository.get_pending_timeout_sessions_by_ids(session_ids)

        if not timed_out_sessions:
            return result

        log.info(
            "Found {} sessions with pending timeout that need termination",
            len(timed_out_sessions),
        )

        # Build session map for getting current status
        session_map = {s.session_info.identity.id: s for s in sessions}

        # Add timed out sessions to failures - Coordinator will apply policy-based transition
        for timed_out in timed_out_sessions:
            session_data = session_map.get(timed_out.session_id)
            if session_data:
                result.failures.append(
                    SessionTransitionInfo(
                        session_id=timed_out.session_id,
                        from_status=session_data.session_info.lifecycle.status,
                        reason="PENDING_TIMEOUT_EXCEEDED",
                        creation_id=timed_out.creation_id,
                        access_key=timed_out.access_key,
                    )
                )

        return result



"""Base class for post-processors."""

from __future__ import annotations

from abc import ABC, abstractmethod
from dataclasses import dataclass, field
from typing import TYPE_CHECKING

from ai.backend.manager.data.kernel.types import KernelStatus
from ai.backend.manager.data.session.types import SessionStatus

if TYPE_CHECKING:
    from ai.backend.manager.sokovan.scheduler.results import (
        KernelExecutionResult,
        SessionExecutionResult,
    )


# =============================================================================
# Session Post-Processor Types
# =============================================================================


@dataclass
class PostProcessorContext:
    """Context passed to session post-processors."""

    result: SessionExecutionResult
    target_statuses: set[SessionStatus] = field(default_factory=set)


class PostProcessor(ABC):
    """Abstract base class for session post-processors.

    Post-processors execute common logic after handler execution.
    They receive the execution result and target status, and perform
    side effects like marking schedules or invalidating caches.
    """

    @abstractmethod
    async def execute(self, context: PostProcessorContext) -> None:
        """Execute the post-processing logic.

        Args:
            context: Post-processor context containing result and target status
        """
        raise NotImplementedError


# =============================================================================
# Kernel Post-Processor Types
# =============================================================================


@dataclass
class KernelPostProcessorContext:
    """Context passed to kernel post-processors."""

    result: KernelExecutionResult
    target_statuses: set[KernelStatus] = field(default_factory=set)


class KernelPostProcessor(ABC):
    """Abstract base class for kernel post-processors.

    Kernel post-processors execute common logic after kernel handler execution.
    They receive the kernel execution result and perform side effects like
    marking schedules for session termination checks.
    """

    @abstractmethod
    async def execute(self, context: KernelPostProcessorContext) -> None:
        """Execute the kernel post-processing logic.

        Args:
            context: Kernel post-processor context containing result
        """
        raise NotImplementedError



"""Types for fair share repository operations.

Contains Scope dataclasses for search operations and entity-based result types.
"""

from __future__ import annotations

import uuid
from collections.abc import Sequence
from dataclasses import dataclass

import sqlalchemy as sa

from ai.backend.manager.data.fair_share import (
    DomainFairShareData,
    ProjectFairShareData,
    UserFairShareData,
)
from ai.backend.manager.errors.resource import (
    DomainNotFound,
    ProjectNotFound,
    ScalingGroupNotFound,
)
from ai.backend.manager.models.domain import DomainRow
from ai.backend.manager.models.group import GroupRow
from ai.backend.manager.models.scaling_group import (
    ScalingGroupForDomainRow,
    ScalingGroupForProjectRow,
    ScalingGroupRow,
)
from ai.backend.manager.repositories.base import ExistenceCheck, QueryCondition, SearchScope

__all__ = (
    # Scope types
    "DomainFairShareSearchScope",
    "ProjectFairShareSearchScope",
    "UserFairShareSearchScope",
    # Entity-based search results
    "DomainFairShareEntitySearchResult",
    "ProjectFairShareEntitySearchResult",
    "UserFairShareEntitySearchResult",
)


# ==================== Scope Types ====================


@dataclass(frozen=True)
class DomainFairShareSearchScope(SearchScope):
    """Required scope for domain fair share entity search.

    Used for Field-level queries where resource_group is determined by parent context.
    """

    resource_group: str
    """Required. The scaling group to search within."""

    def to_condition(self) -> QueryCondition:
        """Convert scope to a query condition for ScalingGroupForDomainRow."""
        resource_group = self.resource_group

        def inner() -> sa.sql.expression.ColumnElement[bool]:
            return ScalingGroupForDomainRow.scaling_group == resource_group

        return inner

    @property
    def existence_checks(self) -> Sequence[ExistenceCheck[str]]:
        """Return existence checks for scope validation."""
        return [
            ExistenceCheck(
                column=ScalingGroupRow.name,
                value=self.resource_group,
                error=ScalingGroupNotFound(self.resource_group),
            ),
        ]


@dataclass(frozen=True)
class ProjectFairShareSearchScope(SearchScope):
    """Required scope for project fair share entity search.

    Used for Field-level queries where resource_group and domain are determined by parent context.
    """

    resource_group: str
    """Required. The scaling group to search within."""

    domain_name: str
    """Required. The domain to search within."""

    def to_condition(self) -> QueryCondition:
        """Convert scope to a query condition for ScalingGroupForProjectRow joined with DomainRow."""
        resource_group = self.resource_group
        domain_name = self.domain_name

        def inner() -> sa.sql.expression.ColumnElement[bool]:
            return sa.and_(
                ScalingGroupForProjectRow.scaling_group == resource_group,
                DomainRow.name == domain_name,
            )

        return inner

    @property
    def existence_checks(self) -> Sequence[ExistenceCheck[str]]:
        """Return existence checks for scope validation."""
        return [
            ExistenceCheck(
                column=ScalingGroupRow.name,
                value=self.resource_group,
                error=ScalingGroupNotFound(self.resource_group),
            ),
            ExistenceCheck(
                column=DomainRow.name,
                value=self.domain_name,
                error=DomainNotFound(self.domain_name),
            ),
        ]


@dataclass(frozen=True)
class UserFairShareSearchScope(SearchScope):
    """Required scope for user fair share entity search.

    Used for Field-level queries where resource_group, domain, and project are determined by parent context.
    """

    resource_group: str
    """Required. The scaling group to search within."""

    domain_name: str
    """Required. The domain to search within."""

    project_id: uuid.UUID
    """Required. The project to search within."""

    def to_condition(self) -> QueryCondition:
        """Convert scope to a query condition for ScalingGroupForProjectRow joined with DomainRow and GroupRow."""
        resource_group = self.resource_group
        domain_name = self.domain_name
        project_id = self.project_id

        def inner() -> sa.sql.expression.ColumnElement[bool]:
            return sa.and_(
                ScalingGroupForProjectRow.scaling_group == resource_group,
                DomainRow.name == domain_name,
                GroupRow.id == project_id,
            )

        return inner

    @property
    def existence_checks(
        self,
    ) -> Sequence[ExistenceCheck[str] | ExistenceCheck[uuid.UUID]]:
        """Return existence checks for scope validation."""
        return [
            ExistenceCheck(
                column=ScalingGroupRow.name,
                value=self.resource_group,
                error=ScalingGroupNotFound(self.resource_group),
            ),
            ExistenceCheck(
                column=DomainRow.name,
                value=self.domain_name,
                error=DomainNotFound(self.domain_name),
            ),
            ExistenceCheck(
                column=GroupRow.id,
                value=self.project_id,
                error=ProjectNotFound(str(self.project_id)),
            ),
        ]


# ==================== Entity-based Search Results ====================


@dataclass(frozen=True)
class DomainFairShareEntitySearchResult:
    """Search result for entity-based domain fair share query.

    Contains domains associated with a resource group,
    regardless of whether they have fair share records.
    """

    items: list[DomainFairShareData]
    """List of domain fair share data."""

    total_count: int
    """Total number of items matching the query (before pagination)."""

    has_next_page: bool
    """Whether there are more items after the current page."""

    has_previous_page: bool
    """Whether there are items before the current page."""


@dataclass(frozen=True)
class ProjectFairShareEntitySearchResult:
    """Search result for entity-based project fair share query.

    Contains projects associated with a resource group,
    regardless of whether they have fair share records.
    """

    items: list[ProjectFairShareData]
    """List of project fair share data."""

    total_count: int
    """Total number of items matching the query (before pagination)."""

    has_next_page: bool
    """Whether there are more items after the current page."""

    has_previous_page: bool
    """Whether there are items before the current page."""


@dataclass(frozen=True)
class UserFairShareEntitySearchResult:
    """Search result for entity-based user fair share query.

    Contains users associated with a resource group,
    regardless of whether they have fair share records.
    """

    items: list[UserFairShareData]
    """List of user fair share data."""

    total_count: int
    """Total number of items matching the query (before pagination)."""

    has_next_page: bool
    """Whether there are more items after the current page."""

    has_previous_page: bool
    """Whether there are items before the current page."""



"""
Round-robin agent selector implementation for sokovan scheduler.

This selector distributes workloads evenly across agents using
a simple round-robin index.
"""

from collections.abc import Sequence

from .selector import (
    AbstractAgentSelector,
    AgentSelectionConfig,
    AgentSelectionCriteria,
    AgentStateTracker,
    ResourceRequirements,
)


class RoundRobinAgentSelector(AbstractAgentSelector):
    """
    Round-robin agent selector that distributes workloads evenly.

    This selector uses a simple index-based approach for round-robin
    selection. Some variance is acceptable.
    """

    def __init__(self, next_index: int = 0) -> None:
        """
        Initialize with the next index to use.

        Args:
            next_index: The index for the next selection
        """
        self.next_index = next_index

    def name(self) -> str:
        """
        Return the selector name for predicates.
        """
        return "RoundRobinAgentSelector"

    def success_message(self) -> str:
        """
        Return a message describing successful agent selection.
        """
        return "Agent selected using round-robin strategy for even workload distribution"

    def select_tracker_by_strategy(
        self,
        trackers: Sequence[AgentStateTracker],
        _resource_req: ResourceRequirements,
        _criteria: AgentSelectionCriteria,
        _config: AgentSelectionConfig,
    ) -> AgentStateTracker:
        """
        Select an agent tracker using round-robin.

        Assumes trackers are already filtered for compatibility.
        The caller should track and update the index after successful allocation.
        """
        # Sort trackers by agent ID for consistent ordering
        sorted_trackers = sorted(trackers, key=lambda tracker: tracker.original_agent.agent_id)

        # Use modulo to wrap around
        selected_index = self.next_index % len(sorted_trackers)

        return sorted_trackers[selected_index]



import asyncio
import logging
from typing import TYPE_CHECKING, Any, Final, Self

import graphene

from ai.backend.common.lock import EtcdLock
from ai.backend.common.utils import deep_merge
from ai.backend.logging import BraceStyleAdapter
from ai.backend.manager.config.unified import ManagerUnifiedConfig
from ai.backend.manager.models.user import UserRole

from .gql_relay import AsyncNode, Connection, ConnectionResolverResult

if TYPE_CHECKING:
    from .schema import GraphQueryContext

_PREFIX: Final[str] = "ai/backend/config"


log = BraceStyleAdapter(logging.getLogger(__spec__.name))


class AvailableServiceNode(graphene.ObjectType):  # type: ignore[misc]
    """
    Available services for configuration.
    Added in 25.8.0.
    """

    class Meta:
        interfaces = (AsyncNode,)
        description = "Available services for configuration. Added in 25.8.0."

    service_variants = graphene.List(
        graphene.String,
        required=True,
        description='Possible values of "Config.service". Added in 25.8.0.',
    )

    async def resolve_component_variants(self, info: graphene.ResolveInfo) -> list[str]:
        return ["manager", "common"]


class AvailableServiceConnection(Connection):
    """Added in 25.8.0."""

    class Meta:
        node = AvailableServiceNode
        description = "Added in 25.8.0."


class ServiceConfigNode(graphene.ObjectType):  # type: ignore[misc]
    """
    Configuration data for a specific service.
    Added in 25.8.0.
    """

    service = graphene.String(
        required=True,
        description="Service name. See AvailableService.service_variants for possible values. Added in 25.8.0.",
    )
    configuration = graphene.JSONString(
        required=True,
        description="Configuration data. Added in 25.8.0.",
    )
    schema = graphene.JSONString(
        required=True,
        description="JSON schema of the configuration. Added in 25.8.0.",
    )

    class Meta:
        interfaces = (AsyncNode,)
        description = "Configuration data for a specific service. Added in 25.8.0."

    @classmethod
    async def load(cls, info: graphene.ResolveInfo, service: str) -> Self:
        ctx: GraphQueryContext = info.context

        def _fallback(x: Any) -> str:
            return str(x)

        unified_config = ctx.config_provider.config.model_dump(
            mode="json", by_alias=True, fallback=_fallback
        )
        unified_config_schema = ctx.config_provider.config.model_json_schema(mode="serialization")

        return cls(
            id=service,
            service=service,
            configuration=unified_config,
            schema=unified_config_schema,
        )

    @classmethod
    async def get_connection(
        cls,
        info: graphene.ResolveInfo,
        services: list[str],
        filter_expr: str | None = None,
        order_expr: str | None = None,
        offset: int | None = None,
        after: str | None = None,
        first: int | None = None,
        before: str | None = None,
        last: int | None = None,
    ) -> ConnectionResolverResult[Self]:
        tasks = [asyncio.create_task(ServiceConfigNode.load(info, svc)) for svc in services]

        result: list[ServiceConfigNode] = []
        # TODO: Propagate errors
        # errors: list = []

        for task in asyncio.as_completed(tasks):
            try:
                node = await task
                result.append(node)
            except Exception as exc:
                log.error(f"Failed to load service config node: {exc}")

        return ConnectionResolverResult(
            node_list=result,
            cursor=None,
            pagination_order=None,
            requested_page_size=None,
            total_count=len(result),
        )


class ServiceConfigConnection(Connection):
    """Added in 25.8.0."""

    class Meta:
        node = ServiceConfigNode
        description = "Added in 25.8.0."


class ModifyServiceConfigNodeInput(graphene.InputObjectType):  # type: ignore[misc]
    """
    Input data for modifying configuration.
    Added in 25.8.0.
    """

    service = graphene.String(
        required=True,
        description="Service name. See AvailableService.service_variants for possible values. Added in 25.8.0.",
    )
    configuration = graphene.JSONString(
        required=True,
        description="Service configuration data to mutate. Added in 25.8.0.",
    )


class ModifyServiceConfigNodePayload(graphene.ObjectType):  # type: ignore[misc]
    """
    Payload for the ModifyServiceConfigNode mutation.
    Added in 25.8.0.
    """

    service_config = graphene.Field(
        ServiceConfigNode,
        required=True,
        description="ServiceConfiguration Node. Added in 25.8.0.",
    )
    allowed_roles = (UserRole.SUPERADMIN,)


class ModifyServiceConfigNode(graphene.Mutation):  # type: ignore[misc]
    allowed_roles = (UserRole.SUPERADMIN,)

    Output = ModifyServiceConfigNodePayload

    class Meta:
        description = "Updates configuration for a given service. Added in 25.8.0."

    class Arguments:
        input = ModifyServiceConfigNodeInput(required=True, description="Added in 25.8.0.")

    @classmethod
    def _get_etcd_prefix_key(cls, service: str) -> str:
        return f"{_PREFIX}/{service}"

    @classmethod
    async def mutate(
        cls,
        root: Any,
        info: graphene.ResolveInfo,
        input: ModifyServiceConfigNodeInput,
    ) -> ModifyServiceConfigNodePayload:
        ctx: GraphQueryContext = info.context

        ETCD_LOCK_KEY = f"{_PREFIX}/lock/modify_service_config"

        # Acquire etcd lock to prevent concurrent modifications
        async with EtcdLock(ETCD_LOCK_KEY, ctx.etcd, timeout=20, lifetime=20) as _:
            merged_raw_unified_config = deep_merge(
                ctx.config_provider.config.model_dump(by_alias=True),
                input.configuration,
            )

            new_config = ManagerUnifiedConfig.model_validate(
                merged_raw_unified_config, by_name=True
            )
            ctx.config_provider.reload(new_config)

            await ctx.etcd.put_prefix(cls._get_etcd_prefix_key(input.service), input.configuration)

        return ModifyServiceConfigNodePayload(
            service_config=ServiceConfigNode.load(info, input.service),
        )



"""Types for scheduling history repository scopes."""

from __future__ import annotations

from dataclasses import dataclass
from typing import Any
from uuid import UUID

from ai.backend.manager.api.gql.base import UUIDEqualMatchSpec
from ai.backend.manager.errors.deployment import EndpointNotFound
from ai.backend.manager.errors.kernel import SessionNotFound
from ai.backend.manager.errors.service import RouteNotFound
from ai.backend.manager.models.endpoint import EndpointRow
from ai.backend.manager.models.routing import RoutingRow
from ai.backend.manager.models.session import SessionRow
from ai.backend.manager.repositories.base import QueryCondition, SearchScope
from ai.backend.manager.repositories.base.types import ExistenceCheck

from .options import (
    DeploymentHistoryConditions,
    RouteHistoryConditions,
    SessionSchedulingHistoryConditions,
)

__all__ = (
    "SessionSchedulingHistorySearchScope",
    "DeploymentHistorySearchScope",
    "RouteHistorySearchScope",
)


# Session Scheduling History Scope


@dataclass(frozen=True)
class SessionSchedulingHistorySearchScope(SearchScope):
    """Scope for session scheduling history search.

    Used for entity-scoped queries where session_id is the scope parameter.
    """

    session_id: UUID
    """Required. The session to search history for."""

    def to_condition(self) -> QueryCondition:
        """Convert scope to a query condition for SessionSchedulingHistoryRow."""
        return SessionSchedulingHistoryConditions.by_session_id_filter(
            UUIDEqualMatchSpec(value=self.session_id, negated=False)
        )

    @property
    def existence_checks(self) -> list[ExistenceCheck[Any]]:
        """Check that the session exists."""
        return [
            ExistenceCheck(
                column=SessionRow.id,
                value=self.session_id,
                error=SessionNotFound(str(self.session_id)),
            ),
        ]


# Deployment History Scope


@dataclass(frozen=True)
class DeploymentHistorySearchScope(SearchScope):
    """Scope for deployment scheduling history search.

    Used for entity-scoped queries where deployment_id is the scope parameter.
    """

    deployment_id: UUID
    """Required. The deployment to search history for."""

    def to_condition(self) -> QueryCondition:
        """Convert scope to a query condition for DeploymentHistoryRow."""
        return DeploymentHistoryConditions.by_deployment_id_filter(
            UUIDEqualMatchSpec(value=self.deployment_id, negated=False)
        )

    @property
    def existence_checks(self) -> list[ExistenceCheck[Any]]:
        """Check that the deployment (endpoint) exists."""
        return [
            ExistenceCheck(
                column=EndpointRow.id,
                value=self.deployment_id,
                error=EndpointNotFound(str(self.deployment_id)),
            ),
        ]


# Route History Scope


@dataclass(frozen=True)
class RouteHistorySearchScope(SearchScope):
    """Scope for route scheduling history search.

    Used for entity-scoped queries where route_id is the scope parameter.
    """

    route_id: UUID
    """Required. The route to search history for."""

    def to_condition(self) -> QueryCondition:
        """Convert scope to a query condition for RouteHistoryRow."""
        return RouteHistoryConditions.by_route_id_filter(
            UUIDEqualMatchSpec(value=self.route_id, negated=False)
        )

    @property
    def existence_checks(self) -> list[ExistenceCheck[Any]]:
        """Check that the route exists."""
        return [
            ExistenceCheck(
                column=RoutingRow.id,
                value=self.route_id,
                error=RouteNotFound(str(self.route_id)),
            ),
        ]



"""Export repository package for report definitions and registry."""

from .db_source import ExportDBSource
from .registry import ExportReportRegistry
from .repositories import ExportRepositories
from .repository import ExportRepository

__all__ = [
    "ExportDBSource",
    "ExportReportRegistry",
    "ExportRepositories",
    "ExportRepository",
]



"""Replica resolver functions."""

from __future__ import annotations

from collections.abc import AsyncGenerator
from uuid import UUID

import strawberry
from strawberry import ID, Info

from ai.backend.manager.api.gql.base import resolve_global_id
from ai.backend.manager.api.gql.deployment.fetcher.replica import fetch_replicas
from ai.backend.manager.api.gql.deployment.types.replica import (
    ModelReplica,
    ModelReplicaConnection,
    ReplicaFilter,
    ReplicaOrderBy,
    ReplicaStatusChangedPayload,
)
from ai.backend.manager.api.gql.types import StrawberryGQLContext
from ai.backend.manager.services.deployment.actions.get_replica_by_id import (
    GetReplicaByIdAction,
)

# Query resolvers


@strawberry.field(description="Added in 25.16.0")  # type: ignore[misc]
async def replicas(
    info: Info[StrawberryGQLContext],
    filter: ReplicaFilter | None = None,
    order_by: list[ReplicaOrderBy] | None = None,
    before: str | None = None,
    after: str | None = None,
    first: int | None = None,
    last: int | None = None,
    limit: int | None = None,
    offset: int | None = None,
) -> ModelReplicaConnection | None:
    """List replicas with optional filtering and pagination."""
    return await fetch_replicas(
        info=info,
        filter=filter,
        order_by=order_by,
        before=before,
        after=after,
        first=first,
        last=last,
        limit=limit,
        offset=offset,
    )


@strawberry.field(description="Added in 25.16.0")  # type: ignore[misc]
async def replica(id: ID, info: Info[StrawberryGQLContext]) -> ModelReplica | None:
    """Get a specific replica by ID."""
    _, replica_id = resolve_global_id(id)
    processor = info.context.processors.deployment
    result = await processor.get_replica_by_id.wait_for_complete(
        GetReplicaByIdAction(replica_id=UUID(replica_id))
    )
    if result.data is None:
        return None
    return ModelReplica.from_dataclass(result.data)


# Subscription resolvers


@strawberry.subscription(description="Added in 25.16.0")  # type: ignore[misc]
async def replica_status_changed(
    revision_id: ID,
) -> AsyncGenerator[ReplicaStatusChangedPayload, None]:
    """Subscribe to replica status changes."""
    # TODO: Implement actual subscription logic using pub/sub system
    raise NotImplementedError("Subscription not implemented")
    yield  # type: ignore[unreachable]  # Makes this an async generator



from __future__ import annotations

from dataclasses import dataclass

from ai.backend.common.data.permission.types import (
    GLOBAL_SCOPE_ID,
    EntityType,
    FieldType,
    OperationType,
    RelationType,
    RoleSource,
    ScopeType,
)
from ai.backend.manager.data.common.types import SearchResult

from .id import ScopeId

# Re-export types for easier access
__all__ = (
    "GLOBAL_SCOPE_ID",
    "EntityType",
    "FieldType",
    "OperationType",
    "RelationType",
    "RoleSource",
    "ScopeData",
    "ScopeListResult",
    "ScopeType",
)


@dataclass(frozen=True)
class ScopeData:
    """Data for a scope."""

    id: ScopeId
    name: str


@dataclass(frozen=True)
class ScopeListResult(SearchResult[ScopeData]):
    """Result of searching scopes."""

    pass



"""User GraphQL types package."""

from .enums import (
    UserRoleEnumGQL,
    UserStatusEnumGQL,
)
from .filters import (
    UserDomainNestedFilter,
    UserFilterGQL,
    UserOrderByGQL,
    UserOrderFieldGQL,
    UserProjectNestedFilter,
    UserRoleEnumFilterGQL,
    UserStatusEnumFilterGQL,
)
from .inputs import (
    BulkCreateUserInputGQL,
    CreateUserInputGQL,
    DeleteUsersInputGQL,
    PurgeUserInputGQL,
    PurgeUsersInputGQL,
    UpdateUserInputGQL,
)
from .nested import (
    EntityTimestampsGQL,
    UserBasicInfoGQL,
    UserContainerSettingsGQL,
    UserOrganizationInfoGQL,
    UserSecurityInfoGQL,
    UserStatusInfoGQL,
)
from .node import (
    UserV2Connection,
    UserV2Edge,
    UserV2GQL,
)
from .payloads import (
    BulkCreateUserErrorGQL,
    BulkCreateUsersPayloadGQL,
    CreateUserPayloadGQL,
    DeleteUserPayloadGQL,
    DeleteUsersPayloadGQL,
    PurgeUserPayloadGQL,
    PurgeUsersPayloadGQL,
    UpdateUserPayloadGQL,
)
from .scopes import (
    DomainUserScopeGQL,
    ProjectUserScopeGQL,
)

__all__ = [
    # Enums
    "UserStatusEnumGQL",
    "UserRoleEnumGQL",
    # Nested Types
    "UserBasicInfoGQL",
    "UserStatusInfoGQL",
    "UserOrganizationInfoGQL",
    "UserSecurityInfoGQL",
    "UserContainerSettingsGQL",
    "EntityTimestampsGQL",
    # Node Types
    "UserV2GQL",
    "UserV2Edge",
    "UserV2Connection",
    # Filters
    "UserDomainNestedFilter",
    "UserProjectNestedFilter",
    "UserFilterGQL",
    "UserOrderFieldGQL",
    "UserOrderByGQL",
    "UserStatusEnumFilterGQL",
    "UserRoleEnumFilterGQL",
    # Scopes
    "DomainUserScopeGQL",
    "ProjectUserScopeGQL",
    # Inputs
    "CreateUserInputGQL",
    "BulkCreateUserInputGQL",
    "UpdateUserInputGQL",
    "DeleteUsersInputGQL",
    "PurgeUserInputGQL",
    "PurgeUsersInputGQL",
    # Payloads
    "CreateUserPayloadGQL",
    "BulkCreateUserErrorGQL",
    "BulkCreateUsersPayloadGQL",
    "UpdateUserPayloadGQL",
    "DeleteUserPayloadGQL",
    "DeleteUsersPayloadGQL",
    "PurgeUserPayloadGQL",
    "PurgeUsersPayloadGQL",
]



from dataclasses import dataclass
from typing import Self

from ai.backend.manager.repositories.resource_preset.repository import ResourcePresetRepository
from ai.backend.manager.repositories.types import RepositoryArgs


@dataclass
class ResourcePresetRepositories:
    repository: ResourcePresetRepository

    @classmethod
    def create(cls, args: RepositoryArgs) -> Self:
        return cls(
            repository=ResourcePresetRepository(
                args.db, args.valkey_stat_client, args.config_provider
            ),
        )



"""Handler for synchronizing healthy routes to service discovery backend."""

import logging
from collections.abc import Sequence

from ai.backend.common.events.dispatcher import EventProducer
from ai.backend.logging import BraceStyleAdapter
from ai.backend.manager.data.deployment.types import RouteStatus, RouteStatusTransitions
from ai.backend.manager.defs import LockID
from ai.backend.manager.repositories.deployment.types import RouteData
from ai.backend.manager.sokovan.deployment.route.executor import RouteExecutor
from ai.backend.manager.sokovan.deployment.route.types import RouteExecutionResult

from .base import RouteHandler

log = BraceStyleAdapter(logging.getLogger(__name__))


class ServiceDiscoverySyncHandler(RouteHandler):
    """Handler for syncing healthy routes to service discovery backend."""

    def __init__(
        self,
        route_executor: RouteExecutor,
        event_producer: EventProducer,
    ) -> None:
        self._route_executor = route_executor
        self._event_producer = event_producer

    @classmethod
    def name(cls) -> str:
        """Get the name of the handler."""
        return "service-discovery-sync"

    @property
    def lock_id(self) -> LockID | None:
        """No lock needed for service discovery sync."""
        return None

    @classmethod
    def target_statuses(cls) -> list[RouteStatus]:
        """Get the target route statuses for this handler."""
        return [RouteStatus.HEALTHY]

    @classmethod
    def next_status(cls) -> RouteStatus | None:
        """No status transition for service discovery sync."""
        return None

    @classmethod
    def failure_status(cls) -> RouteStatus | None:
        """No failure status for service discovery sync."""
        return None

    @classmethod
    def stale_status(cls) -> RouteStatus | None:
        """Get the stale route status if applicable."""
        return None

    @classmethod
    def status_transitions(cls) -> RouteStatusTransitions:
        """Define state transitions for service discovery sync handler (BEP-1030).

        All transitions are None because this handler only syncs to service discovery,
        it doesn't change route status.
        """
        return RouteStatusTransitions(
            success=None,
            failure=None,
            stale=None,
        )

    async def execute(self, routes: Sequence[RouteData]) -> RouteExecutionResult:
        """Execute service discovery synchronization for healthy routes."""
        log.debug("Syncing {} healthy routes to service discovery", len(routes))

        # Execute service discovery sync logic via executor
        return await self._route_executor.sync_service_discovery(routes)

    async def post_process(self, result: RouteExecutionResult) -> None:
        """Handle post-processing after service discovery sync."""
        synced_count = len(result.successes)
        failed_count = len(result.errors)

        if failed_count > 0:
            log.warning(
                "Service discovery sync complete: {} synced, {} failed",
                synced_count,
                failed_count,
            )

            # Log details of failed syncs
            for error in result.errors:
                log.warning(
                    "Failed to sync route {} to service discovery: {}",
                    error.route_info.route_id,
                    error.reason,
                )
        else:
            log.trace("Successfully synced {} routes to service discovery", synced_count)



"""User GraphQL filter and order-by types."""

from __future__ import annotations

from enum import StrEnum
from typing import override

import strawberry

from ai.backend.common.data.user.types import UserRole
from ai.backend.manager.api.gql.base import (
    DateTimeFilter,
    OrderDirection,
    StringFilter,
    UUIDFilter,
)
from ai.backend.manager.api.gql.types import GQLFilter, GQLOrderBy
from ai.backend.manager.data.user.types import UserStatus
from ai.backend.manager.repositories.base import (
    QueryCondition,
    QueryOrder,
    combine_conditions_or,
    negate_conditions,
)
from ai.backend.manager.repositories.domain.options import DomainConditions
from ai.backend.manager.repositories.group.options import GroupConditions
from ai.backend.manager.repositories.user.options import UserConditions, UserOrders

from .enums import UserRoleEnumGQL, UserStatusEnumGQL


@strawberry.input(
    name="UserDomainNestedFilter",
    description=(
        "Added in 26.2.0. Nested filter for the domain a user belongs to. "
        "Filters users whose domain matches all specified conditions."
    ),
)
class UserDomainNestedFilter:
    """Nested filter for domain of a user."""

    name: StringFilter | None = strawberry.field(
        default=None,
        description="Filter by domain name. Supports equals, contains, startsWith, and endsWith.",
    )
    is_active: bool | None = strawberry.field(
        default=None,
        description="Filter by domain active status.",
    )

    def build_conditions(self) -> list[QueryCondition]:
        """Build query conditions for domain nested filter.

        Returns:
            List containing a single EXISTS condition wrapping all domain sub-conditions,
            or empty list if no filters specified.
        """
        raw_conditions: list[QueryCondition] = []
        if self.name:
            condition = self.name.build_query_condition(
                contains_factory=lambda spec: DomainConditions.by_name_contains(spec),
                equals_factory=lambda spec: DomainConditions.by_name_equals(spec),
                starts_with_factory=lambda spec: DomainConditions.by_name_starts_with(spec),
                ends_with_factory=lambda spec: DomainConditions.by_name_ends_with(spec),
            )
            if condition:
                raw_conditions.append(condition)
        if self.is_active is not None:
            raw_conditions.append(DomainConditions.by_is_active(self.is_active))
        if not raw_conditions:
            return []
        return [UserConditions.exists_domain_combined(raw_conditions)]


@strawberry.input(
    name="UserProjectNestedFilter",
    description=(
        "Added in 26.2.0. Nested filter for projects a user belongs to. "
        "Filters users that belong to at least one project matching all specified conditions."
    ),
)
class UserProjectNestedFilter:
    """Nested filter for projects of a user."""

    name: StringFilter | None = strawberry.field(
        default=None,
        description="Filter by project name. Supports equals, contains, startsWith, and endsWith.",
    )
    is_active: bool | None = strawberry.field(
        default=None,
        description="Filter by project active status.",
    )

    def build_conditions(self) -> list[QueryCondition]:
        """Build query conditions for project nested filter.

        Returns:
            List containing a single EXISTS condition wrapping all project sub-conditions,
            or empty list if no filters specified.
        """
        raw_conditions: list[QueryCondition] = []
        if self.name:
            condition = self.name.build_query_condition(
                contains_factory=lambda spec: GroupConditions.by_name_contains(spec),
                equals_factory=lambda spec: GroupConditions.by_name_equals(spec),
                starts_with_factory=lambda spec: GroupConditions.by_name_starts_with(spec),
                ends_with_factory=lambda spec: GroupConditions.by_name_ends_with(spec),
            )
            if condition:
                raw_conditions.append(condition)
        if self.is_active is not None:
            raw_conditions.append(GroupConditions.by_is_active(self.is_active))
        if not raw_conditions:
            return []
        return [UserConditions.exists_project_combined(raw_conditions)]


@strawberry.input(
    name="UserStatusV2EnumFilter",
    description=(
        "Added in 26.2.0. Filter for UserStatusV2 enum fields."
        "Supports equals, in, not_equals, and not_in operations."
    ),
)
class UserStatusEnumFilterGQL:
    """Filter for user status enum fields."""

    equals: UserStatusEnumGQL | None = strawberry.field(
        default=None,
        description="Exact match for user status.",
    )
    in_: list[UserStatusEnumGQL] | None = strawberry.field(
        name="in",
        default=None,
        description="Match any of the provided statuses.",
    )
    not_equals: UserStatusEnumGQL | None = strawberry.field(
        default=None,
        description="Exclude exact status match.",
    )
    not_in: list[UserStatusEnumGQL] | None = strawberry.field(
        default=None,
        description="Exclude any of the provided statuses.",
    )


@strawberry.input(
    name="UserRoleV2EnumFilter",
    description=(
        "Added in 26.2.0. Filter for UserRoleV2 enum fields."
        "Supports equals, in, not_equals, and not_in operations."
    ),
)
class UserRoleEnumFilterGQL:
    """Filter for user role enum fields."""

    equals: UserRoleEnumGQL | None = strawberry.field(
        default=None,
        description="Exact match for user role.",
    )
    in_: list[UserRoleEnumGQL] | None = strawberry.field(
        name="in",
        default=None,
        description="Match any of the provided roles.",
    )
    not_equals: UserRoleEnumGQL | None = strawberry.field(
        default=None,
        description="Exclude exact role match.",
    )
    not_in: list[UserRoleEnumGQL] | None = strawberry.field(
        default=None,
        description="Exclude any of the provided roles.",
    )


@strawberry.input(
    name="UserV2Filter",
    description=(
        "Added in 26.2.0. Filter input for querying users. "
        "Supports filtering by UUID, username, email, status, domain, role, creation time, "
        "and nested domain/project filters. "
        "Multiple filters can be combined using AND, OR, and NOT logical operators."
    ),
)
class UserFilterGQL(GQLFilter):
    """Filter for user queries."""

    uuid: UUIDFilter | None = strawberry.field(
        default=None,
        description="Filter by user UUID. Supports equals and 'in' operations.",
    )
    username: StringFilter | None = strawberry.field(
        default=None,
        description="Filter by username. Supports equals, contains, startsWith, and endsWith.",
    )
    email: StringFilter | None = strawberry.field(
        default=None,
        description="Filter by email address. Supports equals, contains, startsWith, and endsWith.",
    )
    status: UserStatusEnumFilterGQL | None = strawberry.field(
        default=None,
        description="Filter by account status. Supports equals, in, not_equals, and not_in operations.",
    )
    domain_name: StringFilter | None = strawberry.field(
        default=None,
        description="Filter by domain name. Supports equals, contains, startsWith, and endsWith.",
    )
    role: UserRoleEnumFilterGQL | None = strawberry.field(
        default=None,
        description="Filter by user role. Supports equals, in, not_equals, and not_in operations.",
    )
    created_at: DateTimeFilter | None = strawberry.field(
        default=None,
        description="Filter by creation timestamp. Supports before, after, and between operations.",
    )
    domain: UserDomainNestedFilter | None = strawberry.field(
        default=None,
        description=(
            "Filter by nested domain conditions. "
            "Returns users whose domain matches all specified conditions."
        ),
    )
    project: UserProjectNestedFilter | None = strawberry.field(
        default=None,
        description=(
            "Filter by nested project conditions. "
            "Returns users that belong to at least one project matching all specified conditions."
        ),
    )

    AND: list[UserFilterGQL] | None = strawberry.field(
        default=None,
        description="Combine multiple filters with AND logic. All conditions must match.",
    )
    OR: list[UserFilterGQL] | None = strawberry.field(
        default=None,
        description="Combine multiple filters with OR logic. At least one condition must match.",
    )
    NOT: list[UserFilterGQL] | None = strawberry.field(
        default=None,
        description="Negate the specified filters. Records matching these conditions will be excluded.",
    )

    @override
    def build_conditions(self) -> list[QueryCondition]:
        """Build query conditions from filter fields.

        Returns:
            List of QueryCondition callables.
        """
        conditions: list[QueryCondition] = []

        if self.uuid:
            condition = self.uuid.build_query_condition(
                equals_factory=lambda spec: UserConditions.by_uuid_equals(spec),
                in_factory=lambda spec: UserConditions.by_uuid_in(spec),
            )
            if condition:
                conditions.append(condition)

        if self.username:
            condition = self.username.build_query_condition(
                contains_factory=lambda spec: UserConditions.by_username_contains(spec),
                equals_factory=lambda spec: UserConditions.by_username_equals(spec),
                starts_with_factory=lambda spec: UserConditions.by_username_starts_with(spec),
                ends_with_factory=lambda spec: UserConditions.by_username_ends_with(spec),
            )
            if condition:
                conditions.append(condition)

        if self.email:
            condition = self.email.build_query_condition(
                contains_factory=lambda spec: UserConditions.by_email_contains(spec),
                equals_factory=lambda spec: UserConditions.by_email_equals(spec),
                starts_with_factory=lambda spec: UserConditions.by_email_starts_with(spec),
                ends_with_factory=lambda spec: UserConditions.by_email_ends_with(spec),
            )
            if condition:
                conditions.append(condition)

        if self.status:
            if self.status.equals:
                conditions.append(
                    UserConditions.by_status_equals(UserStatus[self.status.equals.name])
                )
            if self.status.in_:
                conditions.append(
                    UserConditions.by_status_in([
                        UserStatus[status.name] for status in self.status.in_
                    ])
                )

        if self.domain_name:
            condition = self.domain_name.build_query_condition(
                contains_factory=lambda spec: UserConditions.by_domain_name_contains(spec),
                equals_factory=lambda spec: UserConditions.by_domain_name_equals(spec),
                starts_with_factory=lambda spec: UserConditions.by_domain_name_starts_with(spec),
                ends_with_factory=lambda spec: UserConditions.by_domain_name_ends_with(spec),
            )
            if condition:
                conditions.append(condition)

        if self.role:
            if self.role.equals:
                conditions.append(UserConditions.by_role_equals(UserRole[self.role.equals.name]))
            if self.role.in_:
                conditions.append(
                    UserConditions.by_role_in([UserRole[role.name] for role in self.role.in_])
                )

        if self.created_at:
            condition = self.created_at.build_query_condition(
                before_factory=lambda dt: UserConditions.by_created_at_before(dt),
                after_factory=lambda dt: UserConditions.by_created_at_after(dt),
            )
            if condition:
                conditions.append(condition)

        if self.domain:
            conditions.extend(self.domain.build_conditions())

        if self.project:
            conditions.extend(self.project.build_conditions())

        # Handle logical operators
        if self.AND:
            for sub_filter in self.AND:
                conditions.extend(sub_filter.build_conditions())

        if self.OR:
            or_sub_conditions: list[QueryCondition] = []
            for sub_filter in self.OR:
                or_sub_conditions.extend(sub_filter.build_conditions())
            if or_sub_conditions:
                conditions.append(combine_conditions_or(or_sub_conditions))

        if self.NOT:
            not_sub_conditions: list[QueryCondition] = []
            for sub_filter in self.NOT:
                not_sub_conditions.extend(sub_filter.build_conditions())
            if not_sub_conditions:
                conditions.append(negate_conditions(not_sub_conditions))

        return conditions


@strawberry.enum(
    name="UserV2OrderField",
    description=(
        "Added in 26.2.0. Fields available for ordering user query results. "
        "CREATED_AT: Order by creation timestamp. "
        "MODIFIED_AT: Order by last modification timestamp. "
        "USERNAME: Order by username alphabetically. "
        "EMAIL: Order by email address alphabetically. "
        "STATUS: Order by account status. "
        "DOMAIN_NAME: Order by domain name (scalar subquery). "
        "PROJECT_NAME: Order by project name (MIN aggregation)."
    ),
)
class UserOrderFieldGQL(StrEnum):
    CREATED_AT = "created_at"
    MODIFIED_AT = "modified_at"
    USERNAME = "username"
    EMAIL = "email"
    STATUS = "status"
    DOMAIN_NAME = "domain_name"
    PROJECT_NAME = "project_name"


@strawberry.input(
    name="UserV2OrderBy",
    description=(
        "Added in 26.2.0. Specifies ordering for user query results. "
        "Combine field selection with direction to sort results. "
        "Default direction is DESC (descending)."
    ),
)
class UserOrderByGQL(GQLOrderBy):
    """OrderBy for user queries."""

    field: UserOrderFieldGQL = strawberry.field(
        description="The field to order by. See UserOrderField for available options."
    )
    direction: OrderDirection = strawberry.field(
        default=OrderDirection.DESC,
        description="Sort direction. ASC for ascending, DESC for descending.",
    )

    @override
    def to_query_order(self) -> QueryOrder:
        """Convert to repository QueryOrder.

        Returns:
            QueryOrder for the specified field and direction.
        """
        ascending = self.direction == OrderDirection.ASC
        match self.field:
            case UserOrderFieldGQL.CREATED_AT:
                return UserOrders.created_at(ascending)
            case UserOrderFieldGQL.MODIFIED_AT:
                return UserOrders.modified_at(ascending)
            case UserOrderFieldGQL.USERNAME:
                return UserOrders.username(ascending)
            case UserOrderFieldGQL.EMAIL:
                return UserOrders.email(ascending)
            case UserOrderFieldGQL.STATUS:
                return UserOrders.status(ascending)
            case UserOrderFieldGQL.DOMAIN_NAME:
                return UserOrders.by_domain_name(ascending)
            case UserOrderFieldGQL.PROJECT_NAME:
                return UserOrders.by_project_name(ascending)
            case _:
                raise ValueError(f"Unknown order field: {self.field}")



"""
Base adapters for converting DTOs to repository query objects.
Provides reusable conversion logic for common patterns.
"""

from __future__ import annotations

from collections.abc import Callable
from typing import final

from ai.backend.common.dto.manager.query import StringFilter, UUIDFilter
from ai.backend.manager.api.gql.base import StringMatchSpec, UUIDEqualMatchSpec, UUIDInMatchSpec
from ai.backend.manager.repositories.base import QueryCondition


class BaseFilterAdapter:
    """Base adapter providing common filter conversion utilities."""

    @final
    def convert_string_filter(
        self,
        string_filter: StringFilter,
        contains_factory: Callable[[StringMatchSpec], QueryCondition],
        equals_factory: Callable[[StringMatchSpec], QueryCondition],
        starts_with_factory: Callable[[StringMatchSpec], QueryCondition],
        ends_with_factory: Callable[[StringMatchSpec], QueryCondition],
    ) -> QueryCondition | None:
        """
        Convert StringFilter to QueryCondition using provided factory callables.

        The method signature matches GraphQL StringFilter.build_query_condition for consistency.

        Args:
            string_filter: The string filter to convert
            contains_factory: Factory for LIKE '%value%' operations
            equals_factory: Factory for exact match (=) operations
            starts_with_factory: Factory for LIKE 'value%' operations
            ends_with_factory: Factory for LIKE '%value' operations

        Returns:
            QueryCondition if any filter field is set, None otherwise
        """
        # equals operations
        if string_filter.equals is not None:
            return equals_factory(
                StringMatchSpec(string_filter.equals, case_insensitive=False, negated=False)
            )
        if string_filter.i_equals is not None:
            return equals_factory(
                StringMatchSpec(string_filter.i_equals, case_insensitive=True, negated=False)
            )
        if string_filter.not_equals is not None:
            return equals_factory(
                StringMatchSpec(string_filter.not_equals, case_insensitive=False, negated=True)
            )
        if string_filter.i_not_equals is not None:
            return equals_factory(
                StringMatchSpec(string_filter.i_not_equals, case_insensitive=True, negated=True)
            )

        # contains operations
        if string_filter.contains is not None:
            return contains_factory(
                StringMatchSpec(string_filter.contains, case_insensitive=False, negated=False)
            )
        if string_filter.i_contains is not None:
            return contains_factory(
                StringMatchSpec(string_filter.i_contains, case_insensitive=True, negated=False)
            )
        if string_filter.not_contains is not None:
            return contains_factory(
                StringMatchSpec(string_filter.not_contains, case_insensitive=False, negated=True)
            )
        if string_filter.i_not_contains is not None:
            return contains_factory(
                StringMatchSpec(string_filter.i_not_contains, case_insensitive=True, negated=True)
            )

        # starts_with operations
        if string_filter.starts_with is not None:
            return starts_with_factory(
                StringMatchSpec(string_filter.starts_with, case_insensitive=False, negated=False)
            )
        if string_filter.i_starts_with is not None:
            return starts_with_factory(
                StringMatchSpec(string_filter.i_starts_with, case_insensitive=True, negated=False)
            )
        if string_filter.not_starts_with is not None:
            return starts_with_factory(
                StringMatchSpec(string_filter.not_starts_with, case_insensitive=False, negated=True)
            )
        if string_filter.i_not_starts_with is not None:
            return starts_with_factory(
                StringMatchSpec(
                    string_filter.i_not_starts_with, case_insensitive=True, negated=True
                )
            )

        # ends_with operations
        if string_filter.ends_with is not None:
            return ends_with_factory(
                StringMatchSpec(string_filter.ends_with, case_insensitive=False, negated=False)
            )
        if string_filter.i_ends_with is not None:
            return ends_with_factory(
                StringMatchSpec(string_filter.i_ends_with, case_insensitive=True, negated=False)
            )
        if string_filter.not_ends_with is not None:
            return ends_with_factory(
                StringMatchSpec(string_filter.not_ends_with, case_insensitive=False, negated=True)
            )
        if string_filter.i_not_ends_with is not None:
            return ends_with_factory(
                StringMatchSpec(string_filter.i_not_ends_with, case_insensitive=True, negated=True)
            )

        return None

    @final
    def convert_uuid_filter(
        self,
        uuid_filter: UUIDFilter,
        equals_factory: Callable[[UUIDEqualMatchSpec], QueryCondition],
        in_factory: Callable[[UUIDInMatchSpec], QueryCondition],
    ) -> QueryCondition | None:
        """
        Convert UUIDFilter to QueryCondition using provided factory callables.

        Args:
            uuid_filter: The UUID filter to convert
            equals_factory: Factory for equality operations (=, !=)
            in_factory: Factory for IN operations (IN, NOT IN)

        Returns:
            QueryCondition if any filter field is set, None otherwise
        """
        # Equality operations
        if uuid_filter.equals is not None:
            return equals_factory(UUIDEqualMatchSpec(value=uuid_filter.equals, negated=False))
        if uuid_filter.not_equals is not None:
            return equals_factory(UUIDEqualMatchSpec(value=uuid_filter.not_equals, negated=True))

        # IN operations
        if uuid_filter.in_ is not None:
            return in_factory(UUIDInMatchSpec(values=uuid_filter.in_, negated=False))
        if uuid_filter.not_in is not None:
            return in_factory(UUIDInMatchSpec(values=uuid_filter.not_in, negated=True))

        return None



"""Action for searching routes of a deployment."""

from __future__ import annotations

from dataclasses import dataclass
from typing import override

from ai.backend.manager.actions.action import BaseActionResult
from ai.backend.manager.actions.types import ActionOperationType
from ai.backend.manager.data.deployment.types import (
    RouteInfo,
)
from ai.backend.manager.repositories.base import BatchQuerier

from .base import RouteBaseAction


@dataclass
class SearchRoutesAction(RouteBaseAction):
    """Action to search routes with filtering and pagination."""

    querier: BatchQuerier

    @override
    def entity_id(self) -> str | None:
        return None

    @override
    @classmethod
    def operation_type(cls) -> ActionOperationType:
        return ActionOperationType.SEARCH


@dataclass
class SearchRoutesActionResult(BaseActionResult):
    """Result of searching routes."""

    routes: list[RouteInfo]
    total_count: int
    has_next_page: bool
    has_previous_page: bool

    @override
    def entity_id(self) -> str | None:
        return None



"""Post-processors for session and kernel handler results."""

from .base import (
    KernelPostProcessor,
    KernelPostProcessorContext,
    PostProcessor,
    PostProcessorContext,
)
from .cache_invalidation import CacheInvalidationPostProcessor
from .factory import create_kernel_post_processors, create_session_post_processors
from .kernel_schedule_marking import KernelScheduleMarkingPostProcessor
from .schedule_marking import ScheduleMarkingPostProcessor

__all__ = [
    # Session post-processors
    "PostProcessor",
    "PostProcessorContext",
    "ScheduleMarkingPostProcessor",
    "CacheInvalidationPostProcessor",
    "create_session_post_processors",
    # Kernel post-processors
    "KernelPostProcessor",
    "KernelPostProcessorContext",
    "KernelScheduleMarkingPostProcessor",
    "create_kernel_post_processors",
]



from __future__ import annotations

import logging
from typing import TYPE_CHECKING

from ai.backend.logging import BraceStyleAdapter
from ai.backend.manager.services.audit_log.actions.create import (
    CreateAuditLogAction,
    CreateAuditLogActionResult,
)
from ai.backend.manager.services.audit_log.actions.search import (
    SearchAuditLogsAction,
    SearchAuditLogsActionResult,
)

if TYPE_CHECKING:
    from ai.backend.manager.repositories.audit_log import AuditLogRepository

log = BraceStyleAdapter(logging.getLogger(__spec__.name))


class AuditLogService:
    _audit_log_repository: AuditLogRepository

    def __init__(self, audit_log_repository: AuditLogRepository) -> None:
        self._audit_log_repository = audit_log_repository

    async def create(self, action: CreateAuditLogAction) -> CreateAuditLogActionResult:
        data = await self._audit_log_repository.create(action.creator)
        return CreateAuditLogActionResult(audit_log_id=data.id)

    async def search(self, action: SearchAuditLogsAction) -> SearchAuditLogsActionResult:
        result = await self._audit_log_repository.search(action.querier)
        return SearchAuditLogsActionResult(
            data=result.items,
            total_count=result.total_count,
            has_next_page=result.has_next_page,
            has_previous_page=result.has_previous_page,
        )



"""Resource Usage GQL types package."""

from .common import (
    UsageBucketMetadataGQL,
    UsageBucketOrderField,
)
from .domain_usage import (
    DomainUsageBucketConnection,
    DomainUsageBucketEdge,
    DomainUsageBucketFilter,
    DomainUsageBucketGQL,
    DomainUsageBucketOrderBy,
)
from .project_usage import (
    ProjectUsageBucketConnection,
    ProjectUsageBucketEdge,
    ProjectUsageBucketFilter,
    ProjectUsageBucketGQL,
    ProjectUsageBucketOrderBy,
)
from .user_usage import (
    UserUsageBucketConnection,
    UserUsageBucketEdge,
    UserUsageBucketFilter,
    UserUsageBucketGQL,
    UserUsageBucketOrderBy,
)

__all__ = [
    # Common
    "UsageBucketMetadataGQL",
    "UsageBucketOrderField",
    # Domain
    "DomainUsageBucketGQL",
    "DomainUsageBucketConnection",
    "DomainUsageBucketEdge",
    "DomainUsageBucketFilter",
    "DomainUsageBucketOrderBy",
    # Project
    "ProjectUsageBucketGQL",
    "ProjectUsageBucketConnection",
    "ProjectUsageBucketEdge",
    "ProjectUsageBucketFilter",
    "ProjectUsageBucketOrderBy",
    # User
    "UserUsageBucketGQL",
    "UserUsageBucketConnection",
    "UserUsageBucketEdge",
    "UserUsageBucketFilter",
    "UserUsageBucketOrderBy",
]



from __future__ import annotations

from .base import SchedulingHistoryAction
from .search_deployment_history import (
    SearchDeploymentHistoryAction,
    SearchDeploymentHistoryActionResult,
)
from .search_deployment_scoped_history import (
    SearchDeploymentScopedHistoryAction,
    SearchDeploymentScopedHistoryActionResult,
)
from .search_route_history import (
    SearchRouteHistoryAction,
    SearchRouteHistoryActionResult,
)
from .search_route_scoped_history import (
    SearchRouteScopedHistoryAction,
    SearchRouteScopedHistoryActionResult,
)
from .search_session_history import (
    SearchSessionHistoryAction,
    SearchSessionHistoryActionResult,
)
from .search_session_scoped_history import (
    SearchSessionScopedHistoryAction,
    SearchSessionScopedHistoryActionResult,
)

__all__ = (
    "SchedulingHistoryAction",
    # Admin actions
    "SearchSessionHistoryAction",
    "SearchSessionHistoryActionResult",
    "SearchDeploymentHistoryAction",
    "SearchDeploymentHistoryActionResult",
    "SearchRouteHistoryAction",
    "SearchRouteHistoryActionResult",
    # Scoped actions (added in 26.2.0)
    "SearchSessionScopedHistoryAction",
    "SearchSessionScopedHistoryActionResult",
    "SearchDeploymentScopedHistoryAction",
    "SearchDeploymentScopedHistoryActionResult",
    "SearchRouteScopedHistoryAction",
    "SearchRouteScopedHistoryActionResult",
)



"""Actions for app configuration service."""

from .base import AppConfigAction
from .domain import (
    DeleteDomainConfigAction,
    DeleteDomainConfigActionResult,
    GetDomainConfigAction,
    GetDomainConfigActionResult,
    UpsertDomainConfigAction,
    UpsertDomainConfigActionResult,
)
from .get_merged import GetMergedAppConfigAction, GetMergedAppConfigActionResult
from .user import (
    DeleteUserConfigAction,
    DeleteUserConfigActionResult,
    GetUserConfigAction,
    GetUserConfigActionResult,
    UpsertUserConfigAction,
    UpsertUserConfigActionResult,
)

__all__ = [
    "AppConfigAction",
    # Domain config actions
    "GetDomainConfigAction",
    "GetDomainConfigActionResult",
    "UpsertDomainConfigAction",
    "UpsertDomainConfigActionResult",
    "DeleteDomainConfigAction",
    "DeleteDomainConfigActionResult",
    # User config actions
    "GetUserConfigAction",
    "GetUserConfigActionResult",
    "UpsertUserConfigAction",
    "UpsertUserConfigActionResult",
    "DeleteUserConfigAction",
    "DeleteUserConfigActionResult",
    # Merged config action
    "GetMergedAppConfigAction",
    "GetMergedAppConfigActionResult",
]



"""Query conditions and orders for domain repository operations."""

from __future__ import annotations

from collections.abc import Collection
from datetime import datetime
from typing import Any

import sqlalchemy as sa

from ai.backend.manager.api.gql.base import StringMatchSpec
from ai.backend.manager.data.user.types import UserStatus
from ai.backend.manager.models.domain.row import DomainRow
from ai.backend.manager.models.group.row import GroupRow
from ai.backend.manager.models.scaling_group import ScalingGroupForDomainRow
from ai.backend.manager.models.user import UserRow
from ai.backend.manager.repositories.base import QueryCondition, QueryOrder

__all__ = (
    "DomainConditions",
    "DomainOrders",
)


class DomainConditions:
    """Query conditions for filtering domains."""

    # ==================== Name Filters ====================

    @staticmethod
    def by_names(names: Collection[str]) -> QueryCondition:
        def inner() -> sa.sql.expression.ColumnElement[bool]:
            return DomainRow.name.in_(names)

        return inner

    @staticmethod
    def by_name_contains(spec: StringMatchSpec) -> QueryCondition:
        def inner() -> sa.sql.expression.ColumnElement[bool]:
            if spec.case_insensitive:
                condition = DomainRow.name.ilike(f"%{spec.value}%")
            else:
                condition = DomainRow.name.like(f"%{spec.value}%")
            if spec.negated:
                condition = sa.not_(condition)
            return condition

        return inner

    @staticmethod
    def by_name_equals(spec: StringMatchSpec) -> QueryCondition:
        def inner() -> sa.sql.expression.ColumnElement[bool]:
            if spec.case_insensitive:
                condition = sa.func.lower(DomainRow.name) == spec.value.lower()
            else:
                condition = DomainRow.name == spec.value
            if spec.negated:
                condition = sa.not_(condition)
            return condition

        return inner

    @staticmethod
    def by_name_starts_with(spec: StringMatchSpec) -> QueryCondition:
        def inner() -> sa.sql.expression.ColumnElement[bool]:
            if spec.case_insensitive:
                condition = DomainRow.name.ilike(f"{spec.value}%")
            else:
                condition = DomainRow.name.like(f"{spec.value}%")
            if spec.negated:
                condition = sa.not_(condition)
            return condition

        return inner

    @staticmethod
    def by_name_ends_with(spec: StringMatchSpec) -> QueryCondition:
        def inner() -> sa.sql.expression.ColumnElement[bool]:
            if spec.case_insensitive:
                condition = DomainRow.name.ilike(f"%{spec.value}")
            else:
                condition = DomainRow.name.like(f"%{spec.value}")
            if spec.negated:
                condition = sa.not_(condition)
            return condition

        return inner

    # ==================== Description Filters ====================

    @staticmethod
    def by_description_contains(spec: StringMatchSpec) -> QueryCondition:
        def inner() -> sa.sql.expression.ColumnElement[bool]:
            if spec.case_insensitive:
                condition = DomainRow.description.ilike(f"%{spec.value}%")
            else:
                condition = DomainRow.description.like(f"%{spec.value}%")
            if spec.negated:
                condition = sa.not_(condition)
            return condition

        return inner

    @staticmethod
    def by_description_equals(spec: StringMatchSpec) -> QueryCondition:
        def inner() -> sa.sql.expression.ColumnElement[bool]:
            if spec.case_insensitive:
                condition = sa.func.lower(DomainRow.description) == spec.value.lower()
            else:
                condition = DomainRow.description == spec.value
            if spec.negated:
                condition = sa.not_(condition)
            return condition

        return inner

    @staticmethod
    def by_description_starts_with(spec: StringMatchSpec) -> QueryCondition:
        def inner() -> sa.sql.expression.ColumnElement[bool]:
            if spec.case_insensitive:
                condition = DomainRow.description.ilike(f"{spec.value}%")
            else:
                condition = DomainRow.description.like(f"{spec.value}%")
            if spec.negated:
                condition = sa.not_(condition)
            return condition

        return inner

    @staticmethod
    def by_description_ends_with(spec: StringMatchSpec) -> QueryCondition:
        def inner() -> sa.sql.expression.ColumnElement[bool]:
            if spec.case_insensitive:
                condition = DomainRow.description.ilike(f"%{spec.value}")
            else:
                condition = DomainRow.description.like(f"%{spec.value}")
            if spec.negated:
                condition = sa.not_(condition)
            return condition

        return inner

    # ==================== Active Status Filters ====================

    @staticmethod
    def by_is_active(is_active: bool) -> QueryCondition:
        def inner() -> sa.sql.expression.ColumnElement[bool]:
            return DomainRow.is_active == is_active

        return inner

    # ==================== Integration ID Filters ====================

    @staticmethod
    def by_integration_id_equals(integration_id: str) -> QueryCondition:
        def inner() -> sa.sql.expression.ColumnElement[bool]:
            return DomainRow.integration_id == integration_id

        return inner

    @staticmethod
    def by_has_integration_id(has_integration: bool) -> QueryCondition:
        def inner() -> sa.sql.expression.ColumnElement[bool]:
            if has_integration:
                return DomainRow.integration_id.is_not(None)
            return DomainRow.integration_id.is_(None)

        return inner

    # ==================== DateTime Filters ====================

    @staticmethod
    def by_created_at_before(dt: datetime) -> QueryCondition:
        """Filter by created_at < datetime."""

        def inner() -> sa.sql.expression.ColumnElement[bool]:
            return DomainRow.created_at < dt

        return inner

    @staticmethod
    def by_created_at_after(dt: datetime) -> QueryCondition:
        """Filter by created_at > datetime."""

        def inner() -> sa.sql.expression.ColumnElement[bool]:
            return DomainRow.created_at > dt

        return inner

    @staticmethod
    def by_modified_at_before(dt: datetime) -> QueryCondition:
        """Filter by modified_at < datetime."""

        def inner() -> sa.sql.expression.ColumnElement[bool]:
            return DomainRow.modified_at < dt

        return inner

    @staticmethod
    def by_modified_at_after(dt: datetime) -> QueryCondition:
        """Filter by modified_at > datetime."""

        def inner() -> sa.sql.expression.ColumnElement[bool]:
            return DomainRow.modified_at > dt

        return inner

    # ==================== Resource Group Filters ====================

    @staticmethod
    def by_resource_group(resource_group: str) -> QueryCondition:
        """Filter domains by resource group (scaling group).

        This requires joining with the sgroups_for_domains table.
        """

        def inner() -> sa.sql.expression.ColumnElement[bool]:
            return sa.exists(
                sa.select(1)
                .where(ScalingGroupForDomainRow.domain == DomainRow.name)
                .where(ScalingGroupForDomainRow.scaling_group == resource_group)
            )

        return inner

    # ==================== Cursor Pagination ====================

    @staticmethod
    def by_cursor_forward(cursor_name: str) -> QueryCondition:
        """Cursor condition for forward pagination (after cursor).

        Uses subquery to get created_at of the cursor row and compare.
        """

        def inner() -> sa.sql.expression.ColumnElement[bool]:
            subquery = (
                sa.select(DomainRow.created_at)
                .where(DomainRow.name == cursor_name)
                .scalar_subquery()
            )
            return DomainRow.created_at < subquery

        return inner

    @staticmethod
    def by_cursor_backward(cursor_name: str) -> QueryCondition:
        """Cursor condition for backward pagination (before cursor).

        Uses subquery to get created_at of the cursor row and compare.
        """

        def inner() -> sa.sql.expression.ColumnElement[bool]:
            subquery = (
                sa.select(DomainRow.created_at)
                .where(DomainRow.name == cursor_name)
                .scalar_subquery()
            )
            return DomainRow.created_at > subquery

        return inner

    # ==================== Project Nested Filters ====================

    @staticmethod
    def _exists_project(
        *project_conditions: sa.sql.expression.ColumnElement[bool],
    ) -> sa.sql.expression.ColumnElement[bool]:
        """EXISTS subquery: Domain → Project (via FK GroupRow.domain_name)."""
        subq = sa.select(sa.literal(1)).where(GroupRow.domain_name == DomainRow.name)
        for cond in project_conditions:
            subq = subq.where(cond)
        return sa.exists(subq)

    @staticmethod
    def by_project_name_contains(spec: StringMatchSpec) -> QueryCondition:
        def inner() -> sa.sql.expression.ColumnElement[bool]:
            if spec.case_insensitive:
                cond = GroupRow.name.ilike(f"%{spec.value}%")
            else:
                cond = GroupRow.name.like(f"%{spec.value}%")
            if spec.negated:
                cond = sa.not_(cond)
            return DomainConditions._exists_project(cond)

        return inner

    @staticmethod
    def by_project_name_equals(spec: StringMatchSpec) -> QueryCondition:
        def inner() -> sa.sql.expression.ColumnElement[bool]:
            if spec.case_insensitive:
                cond = sa.func.lower(GroupRow.name) == spec.value.lower()
            else:
                cond = GroupRow.name == spec.value
            if spec.negated:
                cond = sa.not_(cond)
            return DomainConditions._exists_project(cond)

        return inner

    @staticmethod
    def by_project_is_active(is_active: bool) -> QueryCondition:
        def inner() -> sa.sql.expression.ColumnElement[bool]:
            return DomainConditions._exists_project(GroupRow.is_active == is_active)

        return inner

    @staticmethod
    def exists_project_combined(project_conditions: list[QueryCondition]) -> QueryCondition:
        """Combine multiple project conditions into single EXISTS subquery."""

        def inner() -> sa.sql.expression.ColumnElement[bool]:
            subq = sa.select(sa.literal(1)).where(GroupRow.domain_name == DomainRow.name)
            for cond in project_conditions:
                subq = subq.where(cond())
            return sa.exists(subq)

        return inner

    # ==================== User Nested Filters ====================

    @staticmethod
    def _exists_user(
        *user_conditions: sa.sql.expression.ColumnElement[bool],
    ) -> sa.sql.expression.ColumnElement[bool]:
        """EXISTS subquery: Domain → User (via FK UserRow.domain_name)."""
        subq = sa.select(sa.literal(1)).where(UserRow.domain_name == DomainRow.name)
        for cond in user_conditions:
            subq = subq.where(cond)
        return sa.exists(subq)

    @staticmethod
    def by_user_username_contains(spec: StringMatchSpec) -> QueryCondition:
        def inner() -> sa.sql.expression.ColumnElement[bool]:
            if spec.case_insensitive:
                cond = UserRow.username.ilike(f"%{spec.value}%")
            else:
                cond = UserRow.username.like(f"%{spec.value}%")
            if spec.negated:
                cond = sa.not_(cond)
            return DomainConditions._exists_user(cond)

        return inner

    @staticmethod
    def by_user_username_equals(spec: StringMatchSpec) -> QueryCondition:
        def inner() -> sa.sql.expression.ColumnElement[bool]:
            if spec.case_insensitive:
                cond = sa.func.lower(UserRow.username) == spec.value.lower()
            else:
                cond = UserRow.username == spec.value
            if spec.negated:
                cond = sa.not_(cond)
            return DomainConditions._exists_user(cond)

        return inner

    @staticmethod
    def by_user_email_contains(spec: StringMatchSpec) -> QueryCondition:
        def inner() -> sa.sql.expression.ColumnElement[bool]:
            if spec.case_insensitive:
                cond = UserRow.email.ilike(f"%{spec.value}%")
            else:
                cond = UserRow.email.like(f"%{spec.value}%")
            if spec.negated:
                cond = sa.not_(cond)
            return DomainConditions._exists_user(cond)

        return inner

    @staticmethod
    def by_user_email_equals(spec: StringMatchSpec) -> QueryCondition:
        def inner() -> sa.sql.expression.ColumnElement[bool]:
            if spec.case_insensitive:
                cond = sa.func.lower(UserRow.email) == spec.value.lower()
            else:
                cond = UserRow.email == spec.value
            if spec.negated:
                cond = sa.not_(cond)
            return DomainConditions._exists_user(cond)

        return inner

    @staticmethod
    def by_user_is_active(is_active: bool) -> QueryCondition:
        def inner() -> sa.sql.expression.ColumnElement[bool]:
            if is_active:
                return DomainConditions._exists_user(UserRow.status == UserStatus.ACTIVE)
            return DomainConditions._exists_user(UserRow.status != UserStatus.ACTIVE)

        return inner

    @staticmethod
    def exists_user_combined(user_conditions: list[QueryCondition]) -> QueryCondition:
        """Combine multiple user conditions into single EXISTS subquery."""

        def inner() -> sa.sql.expression.ColumnElement[bool]:
            subq = sa.select(sa.literal(1)).where(UserRow.domain_name == DomainRow.name)
            for cond in user_conditions:
                subq = subq.where(cond())
            return sa.exists(subq)

        return inner


class DomainOrders:
    """Query orders for sorting domains."""

    @staticmethod
    def name(ascending: bool = True) -> QueryOrder:
        if ascending:
            return DomainRow.name.asc()
        return DomainRow.name.desc()

    @staticmethod
    def created_at(ascending: bool = True) -> QueryOrder:
        if ascending:
            return DomainRow.created_at.asc()
        return DomainRow.created_at.desc()

    @staticmethod
    def modified_at(ascending: bool = True) -> QueryOrder:
        if ascending:
            return DomainRow.modified_at.asc()
        return DomainRow.modified_at.desc()

    @staticmethod
    def is_active(ascending: bool = True) -> QueryOrder:
        if ascending:
            return DomainRow.is_active.asc()
        return DomainRow.is_active.desc()

    # ==================== Project Nested Orders ====================

    @staticmethod
    def _scalar_project_min(
        column: sa.ColumnElement[Any] | sa.orm.InstrumentedAttribute[Any],
    ) -> sa.ScalarSelect[Any]:
        """Scalar subquery with MIN for 1:N relationship (Domain → Project)."""
        return (
            sa.select(sa.func.min(column))
            .where(GroupRow.domain_name == DomainRow.name)
            .correlate(DomainRow)
            .scalar_subquery()
        )

    @staticmethod
    def by_project_name(ascending: bool = True) -> QueryOrder:
        subq = DomainOrders._scalar_project_min(GroupRow.name)
        return subq.asc() if ascending else subq.desc()

    # ==================== User Nested Orders ====================

    @staticmethod
    def _scalar_user_min(
        column: sa.ColumnElement[Any] | sa.orm.InstrumentedAttribute[Any],
    ) -> sa.ScalarSelect[Any]:
        """Scalar subquery with MIN for 1:N relationship (Domain → User)."""
        return (
            sa.select(sa.func.min(column))
            .where(UserRow.domain_name == DomainRow.name)
            .correlate(DomainRow)
            .scalar_subquery()
        )

    @staticmethod
    def by_user_username(ascending: bool = True) -> QueryOrder:
        subq = DomainOrders._scalar_user_min(UserRow.username)
        return subq.asc() if ascending else subq.desc()

    @staticmethod
    def by_user_email(ascending: bool = True) -> QueryOrder:
        subq = DomainOrders._scalar_user_min(UserRow.email)
        return subq.asc() if ascending else subq.desc()



"""Kernel observer handlers for the Sokovan scheduler.

Observers collect data from kernels without changing their state.
Unlike handlers that perform status transitions, observers are
read-only operations for metrics, fair share calculation, etc.
"""

from .base import KernelObserver, ObservationResult
from .fair_share import FairShareObserver

__all__ = [
    "KernelObserver",
    "ObservationResult",
    "FairShareObserver",
]



"""Deployment resolver functions."""

from __future__ import annotations

from collections.abc import AsyncGenerator
from uuid import UUID

import strawberry
from strawberry import ID, Info

from ai.backend.manager.api.gql.base import resolve_global_id
from ai.backend.manager.api.gql.deployment.fetcher.deployment import fetch_deployments
from ai.backend.manager.api.gql.deployment.types.deployment import (
    CreateDeploymentInput,
    CreateDeploymentPayload,
    DeleteDeploymentInput,
    DeleteDeploymentPayload,
    DeploymentFilter,
    DeploymentOrderBy,
    DeploymentStatusChangedPayload,
    ModelDeployment,
    ModelDeploymentConnection,
    SyncReplicaInput,
    SyncReplicaPayload,
    UpdateDeploymentInput,
    UpdateDeploymentPayload,
)
from ai.backend.manager.api.gql.types import StrawberryGQLContext
from ai.backend.manager.services.deployment.actions.create_deployment import (
    CreateDeploymentAction,
)
from ai.backend.manager.services.deployment.actions.destroy_deployment import (
    DestroyDeploymentAction,
)
from ai.backend.manager.services.deployment.actions.get_deployment_by_id import (
    GetDeploymentByIdAction,
)
from ai.backend.manager.services.deployment.actions.sync_replicas import SyncReplicaAction
from ai.backend.manager.services.deployment.actions.update_deployment import UpdateDeploymentAction

# Query resolvers


@strawberry.field(description="Added in 25.16.0")  # type: ignore[misc]
async def deployments(
    info: Info[StrawberryGQLContext],
    filter: DeploymentFilter | None = None,
    order_by: list[DeploymentOrderBy] | None = None,
    before: str | None = None,
    after: str | None = None,
    first: int | None = None,
    last: int | None = None,
    limit: int | None = None,
    offset: int | None = None,
) -> ModelDeploymentConnection | None:
    """List deployments with optional filtering and pagination."""
    return await fetch_deployments(
        info=info,
        filter=filter,
        order_by=order_by,
        before=before,
        after=after,
        first=first,
        last=last,
        limit=limit,
        offset=offset,
    )


@strawberry.field(description="Added in 25.16.0")  # type: ignore[misc]
async def deployment(id: ID, info: Info[StrawberryGQLContext]) -> ModelDeployment | None:
    """Get a specific deployment by ID."""
    _, deployment_id = resolve_global_id(id)
    processor = info.context.processors.deployment
    result = await processor.get_deployment_by_id.wait_for_complete(
        GetDeploymentByIdAction(deployment_id=UUID(deployment_id))
    )
    return ModelDeployment.from_dataclass(result.data)


# Mutation resolvers


@strawberry.mutation(description="Added in 25.16.0")  # type: ignore[misc]
async def create_model_deployment(
    input: CreateDeploymentInput, info: Info[StrawberryGQLContext]
) -> CreateDeploymentPayload:
    """Create a new model deployment."""
    processor = info.context.processors.deployment
    result = await processor.create_deployment.wait_for_complete(
        CreateDeploymentAction(creator=input.to_creator())
    )

    return CreateDeploymentPayload(deployment=ModelDeployment.from_dataclass(result.data))


@strawberry.mutation(description="Added in 25.16.0")  # type: ignore[misc]
async def update_model_deployment(
    input: UpdateDeploymentInput, info: Info[StrawberryGQLContext]
) -> UpdateDeploymentPayload:
    """Update an existing model deployment."""
    _, deployment_id = resolve_global_id(input.id)
    processor = info.context.processors.deployment
    action_result = await processor.update_deployment.wait_for_complete(
        UpdateDeploymentAction(updater=input.to_updater(UUID(deployment_id)))
    )
    return UpdateDeploymentPayload(deployment=ModelDeployment.from_dataclass(action_result.data))


@strawberry.mutation(description="Added in 25.16.0")  # type: ignore[misc]
async def delete_model_deployment(
    input: DeleteDeploymentInput, info: Info[StrawberryGQLContext]
) -> DeleteDeploymentPayload:
    """Delete a model deployment."""
    _, deployment_id = resolve_global_id(input.id)
    deployment_processor = info.context.processors.deployment
    _ = await deployment_processor.destroy_deployment.wait_for_complete(
        DestroyDeploymentAction(endpoint_id=UUID(deployment_id))
    )
    return DeleteDeploymentPayload(id=input.id)


@strawberry.mutation(  # type: ignore[misc]
    description="Added in 25.16.0. Force syncs up-to-date replica information. In normal situations this will be automatically handled by Backend.AI schedulers"
)
async def sync_replicas(
    input: SyncReplicaInput, info: Info[StrawberryGQLContext]
) -> SyncReplicaPayload:
    _, deployment_id = resolve_global_id(input.model_deployment_id)
    deployment_processor = info.context.processors.deployment
    await deployment_processor.sync_replicas.wait_for_complete(
        SyncReplicaAction(deployment_id=UUID(deployment_id))
    )
    return SyncReplicaPayload(success=True)


# Subscription resolvers


@strawberry.subscription(description="Added in 25.16.0. Subscribe to deployment status changes")  # type: ignore[misc]
async def deployment_status_changed(
    info: Info[StrawberryGQLContext],
) -> AsyncGenerator[DeploymentStatusChangedPayload, None]:
    """Subscribe to deployment status changes."""
    # TODO: Implement actual subscription logic using pub/sub system
    raise NotImplementedError("Subscription not implemented")
    yield  # type: ignore[unreachable]  # Makes this an async generator



"""
Resource slot domain exceptions.
"""

from __future__ import annotations

from ai.backend.common.exception import (
    BackendAIError,
    ErrorCode,
    ErrorDetail,
    ErrorDomain,
    ErrorOperation,
)


class ResourceSlotTypeNotFound(BackendAIError):
    """Raised when a requested resource slot type does not exist."""

    error_type = "https://api.backend.ai/probs/resource-slot-type-not-found"
    error_title = "Resource slot type not found."

    def error_code(self) -> ErrorCode:
        return ErrorCode(
            domain=ErrorDomain.RESOURCE_PRESET,
            operation=ErrorOperation.READ,
            error_detail=ErrorDetail.NOT_FOUND,
        )


class AgentResourceCapacityExceeded(BackendAIError):
    """Raised when an agent resource update would exceed the slot capacity."""

    error_type = "https://api.backend.ai/probs/agent-resource-capacity-exceeded"
    error_title = "Agent resource capacity exceeded."

    def error_code(self) -> ErrorCode:
        return ErrorCode(
            domain=ErrorDomain.AGENT,
            operation=ErrorOperation.UPDATE,
            error_detail=ErrorDetail.CONFLICT,
        )



from __future__ import annotations

import logging
import uuid
from collections.abc import Sequence
from typing import TYPE_CHECKING, Any, Self, cast

import graphene
import graphql
import sqlalchemy as sa
from graphql import Undefined, UndefinedType

from ai.backend.common.container_registry import AllowedGroupsModel, ContainerRegistryType
from ai.backend.logging import BraceStyleAdapter
from ai.backend.manager.data.container_registry.types import ContainerRegistryData
from ai.backend.manager.defs import PASSWORD_PLACEHOLDER
from ai.backend.manager.models.container_registry import (
    ContainerRegistryRow,
    ContainerRegistryValidator,
    ContainerRegistryValidatorArgs,
)
from ai.backend.manager.models.minilang import FieldSpecItem, OrderSpecItem
from ai.backend.manager.models.minilang.ordering import QueryOrderParser
from ai.backend.manager.models.minilang.queryfilter import QueryFilterParser
from ai.backend.manager.models.rbac import (
    ContainerRegistryScope,
    ProjectScope,
    ScopeType,
    SystemScope,
)
from ai.backend.manager.repositories.base.creator import Creator
from ai.backend.manager.repositories.base.purger import Purger
from ai.backend.manager.repositories.base.updater import Updater
from ai.backend.manager.repositories.container_registry.creators import ContainerRegistryCreatorSpec
from ai.backend.manager.repositories.container_registry.updaters import (
    ContainerRegistryUpdaterSpec,
)
from ai.backend.manager.services.container_registry.actions.create_container_registry import (
    CreateContainerRegistryAction,
)
from ai.backend.manager.services.container_registry.actions.delete_container_registry import (
    DeleteContainerRegistryAction,
)
from ai.backend.manager.services.container_registry.actions.modify_container_registry import (
    ModifyContainerRegistryAction,
)
from ai.backend.manager.types import OptionalState, TriState

from .base import (
    BigInt,
    FilterExprArg,
    OrderExprArg,
    PaginatedConnectionField,
    generate_sql_info_for_gql_connection,
    set_if_set,
)
from .fields import ScopeField
from .gql_relay import AsyncNode, Connection, ConnectionResolverResult
from .group import GroupConnection, GroupNode

if TYPE_CHECKING:
    from .schema import GraphQueryContext
from ai.backend.manager.models.user import UserRole

log = BraceStyleAdapter(logging.getLogger(__spec__.name))

__all__: Sequence[str] = (
    "AllowedGroups",
    "ContainerRegistryNode",
    "ContainerRegistryConnection",
    "ContainerRegistryScopeField",
    "ContainerRegistryTypeField",
    "CreateContainerRegistryNode",
    "ModifyContainerRegistryNode",
    "DeleteContainerRegistryNode",
    # Legacy GraphQL classes (deprecated since 24.09.0)
    "CreateContainerRegistryInput",
    "ModifyContainerRegistryInput",
    "ContainerRegistryConfig",
    "ContainerRegistry",
    "CreateContainerRegistry",
    "ModifyContainerRegistry",
    "DeleteContainerRegistry",
)


class ContainerRegistryTypeField(graphene.Scalar):  # type: ignore[misc]
    """Added in 24.09.0."""

    allowed_values = tuple(t.value for t in ContainerRegistryType)

    @staticmethod
    def serialize(val: ContainerRegistryType) -> str:
        return val.value

    @staticmethod
    def parse_literal(
        node: graphql.language.ast.Node, _variables: dict[str, Any] | None = None
    ) -> ContainerRegistryType | None:
        if isinstance(node, graphql.language.ast.StringValueNode):
            return ContainerRegistryType(node.value)
        return None

    @staticmethod
    def parse_value(value: str) -> ContainerRegistryType:
        return ContainerRegistryType(value)


class ContainerRegistryNode(graphene.ObjectType):  # type: ignore[misc]
    class Meta:
        interfaces = (AsyncNode,)
        description = "Added in 24.09.0."

    row_id = graphene.UUID(
        description="Added in 24.09.0. The UUID type id of DB container_registries row."
    )
    name = graphene.String()
    url = graphene.String(required=True, description="Added in 24.09.0.")
    type = ContainerRegistryTypeField(required=True, description="Added in 24.09.0.")
    registry_name = graphene.String(required=True, description="Added in 24.09.0.")
    is_global = graphene.Boolean(description="Added in 24.09.0.")
    project = graphene.String(description="Added in 24.09.0.")
    username = graphene.String(description="Added in 24.09.0.")
    password = graphene.String(description="Added in 24.09.0.")
    ssl_verify = graphene.Boolean(description="Added in 24.09.0.")
    extra = graphene.JSONString(description="Added in 24.09.3.")
    allowed_groups = PaginatedConnectionField(GroupConnection, description="Added in 25.3.0.")

    _queryfilter_fieldspec: dict[str, FieldSpecItem] = {
        "row_id": ("id", None),
        "registry_name": ("registry_name", None),
    }
    _queryorder_colmap: dict[str, OrderSpecItem] = {
        "row_id": ("id", None),
        "registry_name": ("registry_name", None),
    }

    @classmethod
    def from_dataclass(cls, data: ContainerRegistryData) -> Self:
        return cls(
            id=data.id,  # auto-converted to Relay global ID
            row_id=data.id,
            url=data.url,
            type=data.type,
            registry_name=data.registry_name,
            project=data.project,
            username=data.username,
            password=PASSWORD_PLACEHOLDER if data.password is not None else None,
            ssl_verify=data.ssl_verify,
            is_global=data.is_global,
            extra=data.extra,
        )

    @classmethod
    async def get_node(cls, info: graphene.ResolveInfo, id: str) -> ContainerRegistryNode:
        graph_ctx: GraphQueryContext = info.context
        _, reg_id = AsyncNode.resolve_global_id(info, id)
        select_stmt = sa.select(ContainerRegistryRow).where(ContainerRegistryRow.id == reg_id)
        async with graph_ctx.db.begin_readonly_session() as db_session:
            reg_row = cast(ContainerRegistryRow | None, await db_session.scalar(select_stmt))
            if reg_row is None:
                raise ValueError(f"Container registry not found (id: {reg_id})")
            return cls.from_row(graph_ctx, reg_row)

    @classmethod
    async def get_connection(
        cls,
        info: graphene.ResolveInfo,
        filter_expr: str | None = None,
        order_expr: str | None = None,
        offset: int | None = None,
        after: str | None = None,
        first: int | None = None,
        before: str | None = None,
        last: int | None = None,
    ) -> ConnectionResolverResult[ContainerRegistryNode]:
        graph_ctx: GraphQueryContext = info.context
        _filter_arg = (
            FilterExprArg(filter_expr, QueryFilterParser(cls._queryfilter_fieldspec))
            if filter_expr is not None
            else None
        )
        _order_expr = (
            OrderExprArg(order_expr, QueryOrderParser(cls._queryorder_colmap))
            if order_expr is not None
            else None
        )
        (
            query,
            cnt_query,
            _,
            cursor,
            pagination_order,
            page_size,
        ) = generate_sql_info_for_gql_connection(
            info,
            ContainerRegistryRow,
            ContainerRegistryRow.id,
            _filter_arg,
            _order_expr,
            offset,
            after=after,
            first=first,
            before=before,
            last=last,
        )
        async with graph_ctx.db.begin_readonly_session() as db_session:
            reg_rows = await db_session.scalars(query)
            total_cnt = await db_session.scalar(cnt_query)
        result = [cls.from_row(graph_ctx, cast(ContainerRegistryRow, row)) for row in reg_rows]
        return ConnectionResolverResult(result, cursor, pagination_order, page_size, total_cnt)

    @classmethod
    def from_row(cls, ctx: GraphQueryContext, row: ContainerRegistryRow) -> ContainerRegistryNode:
        return cls(
            id=row.id,  # auto-converted to Relay global ID
            row_id=row.id,
            url=row.url,
            type=row.type,
            registry_name=row.registry_name,
            project=row.project,
            username=row.username,
            password=PASSWORD_PLACEHOLDER if row.password is not None else None,
            ssl_verify=row.ssl_verify,
            is_global=row.is_global,
            extra=row.extra,
        )

    async def resolve_allowed_groups(
        self,
        info: graphene.ResolveInfo,
        filter: str | None = None,
        order: str | None = None,
        offset: int | None = None,
        after: str | None = None,
        first: int | None = None,
        before: str | None = None,
        last: int | None = None,
    ) -> ConnectionResolverResult[GroupNode]:
        scope = SystemScope()

        if self.is_global:
            container_registry_scope = None
        else:
            container_registry_scope = ContainerRegistryScope.parse(f"container_registry:{self.id}")

        return await GroupNode.get_connection(
            info,
            scope,
            container_registry_scope,
            filter_expr=filter,
            order_expr=order,
            offset=offset,
            after=after,
            first=first,
            before=before,
            last=last,
        )


class ContainerRegistryConnection(Connection):
    """Added in 25.3.0."""

    class Meta:
        node = ContainerRegistryNode
        description = "Added in 24.09.0."


class ContainerRegistryScopeField(graphene.Scalar):  # type: ignore[misc]
    class Meta:
        description = "Added in 25.3.0."

    @staticmethod
    def serialize(val: ContainerRegistryScope) -> str:
        if isinstance(val, ContainerRegistryScope):
            return str(val)
        raise ValueError("Invalid ContainerRegistryScope")

    @staticmethod
    def parse_value(value: str) -> ContainerRegistryScope:
        if isinstance(value, str):
            try:
                return ContainerRegistryScope.parse(value)
            except Exception as e:
                raise ValueError(f"Invalid ContainerRegistryScope: {e}") from e
        raise ValueError("Invalid ContainerRegistryScope")

    @staticmethod
    def parse_literal(node: graphql.language.ast.Node) -> ContainerRegistryScope | None:
        if isinstance(node, graphql.language.ast.StringValueNode):
            try:
                return ContainerRegistryScope.parse(node.value)
            except Exception as e:
                raise ValueError(f"Invalid ContainerRegistryScope: {e}") from e
        return None


class AllowedGroups(graphene.InputObjectType):  # type: ignore[misc]
    """
    Added in 25.3.0.
    """

    add = graphene.List(
        graphene.String,
        default_value=[],
        description="List of group_ids to add associations. Added in 25.3.0.",
    )
    remove = graphene.List(
        graphene.String,
        default_value=[],
        description="List of group_ids to remove associations. Added in 25.3.0.",
    )

    def to_model(self) -> AllowedGroupsModel:
        return AllowedGroupsModel(
            add=[] if (self.add is Undefined or self.add is None) else self.add,
            remove=[] if (self.remove is Undefined or self.remove is None) else self.remove,
        )


class CreateContainerRegistryNode(graphene.Mutation):  # type: ignore[misc]
    """
    Deprecated since 25.3.0. use `CreateContainerRegistryNodeV2` instead
    """

    class Meta:
        description = "Added in 24.09.0."

    allowed_roles = (UserRole.SUPERADMIN,)
    container_registry = graphene.Field(ContainerRegistryNode)

    class Arguments:
        url = graphene.String(required=True, description="Added in 24.09.0.")
        type = ContainerRegistryTypeField(
            required=True,
            description=f"Added in 24.09.0. Container Registry type. One of {ContainerRegistryTypeField.allowed_values}.",
        )
        registry_name = graphene.String(required=True, description="Added in 24.09.0.")
        is_global = graphene.Boolean(description="Added in 24.09.0.")
        project = graphene.String(description="Added in 24.09.0.")
        username = graphene.String(description="Added in 24.09.0.")
        password = graphene.String(description="Added in 24.09.0.")
        ssl_verify = graphene.Boolean(description="Added in 24.09.0.")
        extra = graphene.JSONString(description="Added in 24.09.3.")

    @classmethod
    async def mutate(
        cls,
        root: Any,
        info: graphene.ResolveInfo,
        url: str,
        type: ContainerRegistryType,
        registry_name: str,
        is_global: bool | UndefinedType = Undefined,
        project: str | UndefinedType = Undefined,
        username: str | UndefinedType = Undefined,
        password: str | UndefinedType = Undefined,
        ssl_verify: bool | UndefinedType = Undefined,
        extra: dict[str, Any] | UndefinedType = Undefined,
    ) -> CreateContainerRegistryNode:
        ctx: GraphQueryContext = info.context
        validator = ContainerRegistryValidator(
            ContainerRegistryValidatorArgs(
                url=url,
                type=type,
                project=cast(str | None, project if project is not Undefined else None),
            )
        )

        validator.validate()

        def value_or_none(val: Any) -> Any | None:
            return None if val is Undefined else val

        action = CreateContainerRegistryAction(
            creator=Creator(
                spec=ContainerRegistryCreatorSpec(
                    url=url,
                    type=type,
                    registry_name=registry_name,
                    is_global=value_or_none(is_global),
                    project=value_or_none(project),
                    username=value_or_none(username),
                    password=value_or_none(password),
                    ssl_verify=value_or_none(ssl_verify),
                    extra=value_or_none(extra),
                    allowed_groups=None,  # allowed groups are not supported in v1 mutation
                )
            )
        )

        result = (
            await ctx.processors.container_registry.create_container_registry.wait_for_complete(
                action
            )
        )

        return cls(
            container_registry=ContainerRegistryNode.from_dataclass(result.data),
        )


class ModifyContainerRegistryNode(graphene.Mutation):  # type: ignore[misc]
    """
    Deprecated since 25.3.0. use `ModifyContainerRegistryNodeV2` instead
    """

    allowed_roles = (UserRole.SUPERADMIN,)
    container_registry = graphene.Field(ContainerRegistryNode)

    class Meta:
        description = "Added in 24.09.0."

    class Arguments:
        id = graphene.String(
            required=True,
            description="Object id. Can be either global id or object id. Added in 24.09.0.",
        )
        url = graphene.String(description="Added in 24.09.0.")
        type = ContainerRegistryTypeField(
            description=f"Container Registry type. One of {ContainerRegistryTypeField.allowed_values}. Added in 24.09.0."
        )
        registry_name = graphene.String(description="Added in 24.09.0.")
        is_global = graphene.Boolean(description="Added in 24.09.0.")
        project = graphene.String(description="Added in 24.09.0.")
        username = graphene.String(description="Added in 24.09.0.")
        password = graphene.String(description="Added in 24.09.0.")
        ssl_verify = graphene.Boolean(description="Added in 24.09.0.")
        extra = graphene.JSONString(description="Added in 24.09.3.")

    @classmethod
    async def mutate(
        cls,
        root: Any,
        info: graphene.ResolveInfo,
        id: str,
        url: str | UndefinedType = Undefined,
        type: ContainerRegistryType | UndefinedType = Undefined,
        registry_name: str | UndefinedType = Undefined,
        is_global: bool | UndefinedType = Undefined,
        project: str | UndefinedType = Undefined,
        username: str | UndefinedType = Undefined,
        password: str | UndefinedType = Undefined,
        ssl_verify: bool | UndefinedType = Undefined,
        extra: dict[str, Any] | UndefinedType = Undefined,
    ) -> ModifyContainerRegistryNode:
        ctx: GraphQueryContext = info.context

        _, _id = AsyncNode.resolve_global_id(info, id)
        reg_id = uuid.UUID(_id) if _id else uuid.UUID(id)

        action = ModifyContainerRegistryAction(
            updater=Updater(
                spec=ContainerRegistryUpdaterSpec(
                    url=OptionalState.from_graphql(url),
                    type=OptionalState.from_graphql(type),
                    registry_name=OptionalState.from_graphql(registry_name),
                    is_global=TriState.from_graphql(is_global),
                    project=TriState.from_graphql(project),
                    username=TriState.from_graphql(username),
                    password=TriState.from_graphql(password),
                    ssl_verify=TriState.from_graphql(ssl_verify),
                    extra=TriState.from_graphql(extra),
                    allowed_groups=TriState.nop(),  # Not handled in this deprecated mutation
                ),
                pk_value=reg_id,
            )
        )

        # Execute action through processor
        result = (
            await ctx.processors.container_registry.modify_container_registry.wait_for_complete(
                action
            )
        )

        return cls(container_registry=ContainerRegistryNode.from_dataclass(result.data))


class DeleteContainerRegistryNode(graphene.Mutation):  # type: ignore[misc]
    """
    Deprecated since 25.3.0. use `DeleteContainerRegistryNodeV2` instead
    """

    allowed_roles = (UserRole.SUPERADMIN,)
    container_registry = graphene.Field(ContainerRegistryNode)

    class Meta:
        description = "Added in 24.09.0."

    class Arguments:
        id = graphene.String(
            required=True,
            description="Object id. Can be either global id or object id. Added in 24.09.0.",
        )

    @classmethod
    async def mutate(
        cls,
        root: Any,
        info: graphene.ResolveInfo,
        id: str,
    ) -> DeleteContainerRegistryNode:
        ctx: GraphQueryContext = info.context

        _, _id = AsyncNode.resolve_global_id(info, id)
        reg_id = uuid.UUID(_id) if _id else uuid.UUID(id)

        result = (
            await ctx.processors.container_registry.delete_container_registry.wait_for_complete(
                DeleteContainerRegistryAction(
                    purger=Purger(row_class=ContainerRegistryRow, pk_value=reg_id)
                )
            )
        )
        return cls(container_registry=ContainerRegistryNode.from_dataclass(result.data))


class CreateContainerRegistryQuota(graphene.Mutation):  # type: ignore[misc]
    """Added in 25.3.0."""

    allowed_roles = (
        UserRole.SUPERADMIN,
        UserRole.ADMIN,
    )

    class Arguments:
        scope_id = ScopeField(required=True)
        quota = BigInt(required=True)

    ok = graphene.Boolean()
    msg = graphene.String()

    @classmethod
    async def mutate(
        cls,
        root: Any,
        info: graphene.ResolveInfo,
        scope_id: ScopeType,
        quota: int | float,
    ) -> Self:
        graph_ctx: GraphQueryContext = info.context
        try:
            match scope_id:
                case ProjectScope():
                    await (
                        graph_ctx.services_ctx.per_project_container_registries_quota.create_quota(
                            scope_id, int(quota)
                        )
                    )
                case _:
                    raise NotImplementedError("Only project scope is supported for now.")

            return cls(ok=True, msg="success")
        except Exception as e:
            return cls(ok=False, msg=str(e))


class UpdateContainerRegistryQuota(graphene.Mutation):  # type: ignore[misc]
    """Added in 25.3.0."""

    allowed_roles = (
        UserRole.SUPERADMIN,
        UserRole.ADMIN,
    )

    class Arguments:
        scope_id = ScopeField(required=True)
        quota = BigInt(required=True)

    ok = graphene.Boolean()
    msg = graphene.String()

    @classmethod
    async def mutate(
        cls,
        root: Any,
        info: graphene.ResolveInfo,
        scope_id: ScopeType,
        quota: int | float,
    ) -> Self:
        graph_ctx: GraphQueryContext = info.context
        try:
            match scope_id:
                case ProjectScope(_):
                    await (
                        graph_ctx.services_ctx.per_project_container_registries_quota.update_quota(
                            scope_id, int(quota)
                        )
                    )
                case _:
                    raise NotImplementedError("Only project scope is supported for now.")

            return cls(ok=True, msg="success")
        except Exception as e:
            return cls(ok=False, msg=str(e))


class DeleteContainerRegistryQuota(graphene.Mutation):  # type: ignore[misc]
    """Added in 25.3.0."""

    allowed_roles = (
        UserRole.SUPERADMIN,
        UserRole.ADMIN,
    )

    class Arguments:
        scope_id = ScopeField(required=True)

    ok = graphene.Boolean()
    msg = graphene.String()

    @classmethod
    async def mutate(
        cls,
        root: Any,
        info: graphene.ResolveInfo,
        scope_id: ScopeType,
    ) -> Self:
        graph_ctx: GraphQueryContext = info.context
        try:
            match scope_id:
                case ProjectScope(_):
                    await (
                        graph_ctx.services_ctx.per_project_container_registries_quota.delete_quota(
                            scope_id
                        )
                    )
                case _:
                    raise NotImplementedError("Only project scope is supported for now.")

            return cls(ok=True, msg="success")
        except Exception as e:
            return cls(ok=False, msg=str(e))


# ================================================================================
# Legacy GraphQL classes (deprecated since 24.09.0)
# Moved from models/container_registry/row.py
# ================================================================================


class CreateContainerRegistryInput(graphene.InputObjectType):  # type: ignore[misc]
    """
    Deprecated since 24.09.0.
    """

    url = graphene.String(required=True)
    type = graphene.String(required=True)
    project = graphene.List(graphene.String)
    username = graphene.String()
    password = graphene.String()
    ssl_verify = graphene.Boolean()
    is_global = graphene.Boolean(description="Added in 24.09.0.")


class ModifyContainerRegistryInput(graphene.InputObjectType):  # type: ignore[misc]
    """
    Deprecated since 24.09.0.
    """

    url = graphene.String()
    type = graphene.String()
    project = graphene.List(graphene.String)
    username = graphene.String()
    password = graphene.String()
    ssl_verify = graphene.Boolean()
    is_global = graphene.Boolean(description="Added in 24.09.0.")


class ContainerRegistryConfig(graphene.ObjectType):  # type: ignore[misc]
    """
    Deprecated since 24.09.0.
    """

    url = graphene.String(required=True)
    type = graphene.String(required=True)
    project = graphene.List(graphene.String)
    username = graphene.String()
    password = graphene.String()
    ssl_verify = graphene.Boolean()
    is_global = graphene.Boolean(description="Added in 24.09.0.")


class ContainerRegistry(graphene.ObjectType):  # type: ignore[misc]
    """
    Deprecated since 24.09.0. use `ContainerRegistryNode` instead
    """

    hostname = graphene.String()
    config = graphene.Field(ContainerRegistryConfig)

    class Meta:
        interfaces = (AsyncNode,)

    @classmethod
    def from_row(cls, ctx: GraphQueryContext, row: ContainerRegistryRow) -> ContainerRegistry:
        return cls(
            id=row.id,  # auto-converted to Relay global ID
            hostname=row.registry_name,
            config=ContainerRegistryConfig(
                url=row.url,
                type=str(row.type),
                project=[row.project],
                username=row.username,
                password=PASSWORD_PLACEHOLDER if row.password is not None else None,
                ssl_verify=row.ssl_verify,
                is_global=row.is_global,
            ),
        )

    @classmethod
    async def load_by_hostname(cls, ctx: GraphQueryContext, hostname: str) -> ContainerRegistry:
        async with ctx.db.begin_readonly_session() as session:
            return cls.from_row(
                ctx,
                (
                    await ContainerRegistryRow.list_by_registry_name(
                        session,
                        hostname,
                    )
                )[0],
            )

    @classmethod
    async def load_all(
        cls,
        ctx: GraphQueryContext,
    ) -> Sequence[ContainerRegistry]:
        async with ctx.db.begin_readonly_session() as session:
            rows = await session.scalars(sa.select(ContainerRegistryRow))
            return [cls.from_row(ctx, row) for row in rows]


class CreateContainerRegistry(graphene.Mutation):  # type: ignore[misc]
    """
    Deprecated since 24.09.0. use `CreateContainerRegistryNode` instead
    """

    allowed_roles = (UserRole.SUPERADMIN,)
    container_registry = graphene.Field(ContainerRegistry)

    class Arguments:
        hostname = graphene.String(required=True)
        props = CreateContainerRegistryInput(required=True)

    @classmethod
    async def mutate(
        cls,
        root: Any,
        info: graphene.ResolveInfo,
        hostname: str,
        props: CreateContainerRegistryInput,
    ) -> CreateContainerRegistry:
        ctx: GraphQueryContext = info.context

        input_config: dict[str, Any] = {
            "registry_name": hostname,
            "url": props.url,
            "type": ContainerRegistryType(props.type),
        }

        if props.project:
            input_config["project"] = props.project[0]

        set_if_set(props, input_config, "username")
        set_if_set(props, input_config, "password")
        set_if_set(props, input_config, "ssl_verify")
        set_if_set(props, input_config, "is_global")

        async with ctx.db.begin_session() as db_session:
            reg_row = ContainerRegistryRow(id=uuid.uuid4(), **input_config)
            db_session.add(reg_row)
            await db_session.flush()
            await db_session.refresh(reg_row)

            return cls(
                container_registry=ContainerRegistry.from_row(ctx, reg_row),
            )


class ModifyContainerRegistry(graphene.Mutation):  # type: ignore[misc]
    """
    Deprecated since 24.09.0. use `ModifyContainerRegistryNode` instead
    """

    allowed_roles = (UserRole.SUPERADMIN,)
    container_registry = graphene.Field(ContainerRegistry)

    class Arguments:
        hostname = graphene.String(required=True)
        props = ModifyContainerRegistryInput(required=True)

    @classmethod
    async def mutate(
        cls,
        root: Any,
        info: graphene.ResolveInfo,
        hostname: str,
        props: ModifyContainerRegistryInput,
    ) -> ModifyContainerRegistry:
        ctx: GraphQueryContext = info.context

        input_config: dict[str, Any] = {
            "registry_name": hostname,
        }

        if props.project:
            input_config["project"] = props.project[0]

        if props.type:
            input_config["type"] = ContainerRegistryType(props.type)

        set_if_set(props, input_config, "url")
        set_if_set(props, input_config, "is_global")
        set_if_set(props, input_config, "username")
        set_if_set(props, input_config, "password")
        set_if_set(props, input_config, "ssl_verify")

        async with ctx.db.begin_session() as session:
            stmt = sa.select(ContainerRegistryRow).where(
                ContainerRegistryRow.registry_name == hostname
            )
            reg_row = await session.scalar(stmt)
            if reg_row is None:
                raise ValueError(f"ContainerRegistry not found (hostname: {hostname})")

            for field, val in input_config.items():
                setattr(reg_row, field, val)

            return cls(container_registry=ContainerRegistry.from_row(ctx, reg_row))


class DeleteContainerRegistry(graphene.Mutation):  # type: ignore[misc]
    """
    Deprecated since 24.09.0. use `DeleteContainerRegistryNode` instead
    """

    allowed_roles = (UserRole.SUPERADMIN,)
    container_registry = graphene.Field(ContainerRegistry)

    class Arguments:
        hostname = graphene.String(required=True)

    @classmethod
    async def mutate(
        cls,
        root: Any,
        info: graphene.ResolveInfo,
        hostname: str,
    ) -> DeleteContainerRegistry:
        ctx: GraphQueryContext = info.context
        container_registry = await ContainerRegistry.load_by_hostname(ctx, hostname)
        async with ctx.db.begin_session() as session:
            stmt = sa.delete(ContainerRegistryRow).where(
                ContainerRegistryRow.registry_name == hostname
            )
            await session.execute(stmt)
        return cls(container_registry=container_registry)



"""Merge 51dd and d2aa

Revision ID: d643752544de
Revises: 51dddd79aa21, d2aafa234374
Create Date: 2020-03-09 12:04:27.013567

"""

# revision identifiers, used by Alembic.
revision = "d643752544de"
down_revision = ("51dddd79aa21", "d2aafa234374")
branch_labels = None
depends_on = None


def upgrade() -> None:
    pass


def downgrade() -> None:
    pass



from dataclasses import dataclass
from typing import override

from ai.backend.manager.actions.action import BaseActionResult
from ai.backend.manager.actions.types import ActionOperationType
from ai.backend.manager.data.domain.types import (
    DomainData,
    UserInfo,
)
from ai.backend.manager.models.domain import DomainRow
from ai.backend.manager.repositories.base.creator import Creator
from ai.backend.manager.services.domain.actions.base import DomainAction


@dataclass
class CreateDomainNodeAction(DomainAction):
    user_info: UserInfo
    creator: Creator[DomainRow]
    scaling_groups: list[str] | None = None

    @override
    def entity_id(self) -> str | None:
        return None

    @override
    @classmethod
    def operation_type(cls) -> ActionOperationType:
        return ActionOperationType.CREATE


@dataclass
class CreateDomainNodeActionResult(BaseActionResult):
    domain_data: DomainData

    @override
    def entity_id(self) -> str | None:
        return self.domain_data.name



from ai.backend.common.container_registry import ContainerRegistryType

from .row import (
    ContainerRegistryRow,
    ContainerRegistryValidator,
    ContainerRegistryValidatorArgs,
)

__all__ = (
    "ContainerRegistryRow",
    "ContainerRegistryType",
    "ContainerRegistryValidator",
    "ContainerRegistryValidatorArgs",
)



"""Session export report definition."""

from __future__ import annotations

import json
from typing import Any

import sqlalchemy as sa

from ai.backend.manager.defs import DEFAULT_ROLE
from ai.backend.manager.models.group import GroupRow
from ai.backend.manager.models.kernel import KernelRow
from ai.backend.manager.models.resource_policy import ProjectResourcePolicyRow
from ai.backend.manager.models.scaling_group import ScalingGroupRow
from ai.backend.manager.models.session import SessionRow
from ai.backend.manager.models.user import UserRow
from ai.backend.manager.repositories.base.export import (
    ExportFieldDef,
    ExportFieldType,
    JoinDef,
    ReportDef,
)

# =============================================================================
# Helper Functions
# =============================================================================


def _serialize_json(value: Any) -> str:
    """Serialize value to JSON string, converting sets to lists."""
    if not value:
        return ""

    def convert(obj: Any) -> Any:
        if isinstance(obj, set):
            return sorted(obj)
        if isinstance(obj, dict):
            return {k: convert(v) for k, v in obj.items()}
        if isinstance(obj, list):
            return [convert(i) for i in obj]
        return obj

    return json.dumps(convert(dict(value)))


# =============================================================================
# JOIN Definitions
# =============================================================================

# Project JOIN (N:1, no duplication)
PROJECT_JOIN = JoinDef(
    table=GroupRow.__table__,
    condition=SessionRow.group_id == GroupRow.id,
)

# Project Resource Policy JOIN (N:1 through Project, no duplication)
PROJECT_RESOURCE_POLICY_JOIN = JoinDef(
    table=ProjectResourcePolicyRow.__table__,
    condition=GroupRow.resource_policy == ProjectResourcePolicyRow.name,
)
PROJECT_POLICY_JOINS = (PROJECT_JOIN, PROJECT_RESOURCE_POLICY_JOIN)

# User JOIN (N:1, no duplication)
USER_JOIN = JoinDef(
    table=UserRow.__table__,
    condition=SessionRow.user_uuid == UserRow.uuid,
)

# All Kernels JOIN (1:N, causes duplication)
KERNEL_JOIN = JoinDef(
    table=KernelRow.__table__,
    condition=SessionRow.id == KernelRow.session_id,
)

# Main Kernel JOIN (N:1, no duplication - main kernel only)
MAIN_KERNEL_JOIN = JoinDef(
    table=KernelRow.__table__,
    condition=sa.and_(
        SessionRow.id == KernelRow.session_id,
        KernelRow.cluster_role == DEFAULT_ROLE,
    ),
)

# Scaling Group JOIN (N:1, no duplication)
SCALING_GROUP_JOIN = JoinDef(
    table=ScalingGroupRow.__table__,
    condition=SessionRow.scaling_group_name == ScalingGroupRow.name,
)

# Field definitions for session export
SESSION_FIELDS: list[ExportFieldDef] = [
    ExportFieldDef(
        key="id",
        name="ID",
        description="Session UUID",
        field_type=ExportFieldType.UUID,
        column=SessionRow.id,
    ),
    ExportFieldDef(
        key="name",
        name="Name",
        description="Session name",
        field_type=ExportFieldType.STRING,
        column=SessionRow.name,
    ),
    ExportFieldDef(
        key="session_type",
        name="Type",
        description="Session type",
        field_type=ExportFieldType.ENUM,
        column=SessionRow.session_type,
        formatter=lambda v: str(v) if v else "",
    ),
    ExportFieldDef(
        key="domain_name",
        name="Domain",
        description="Domain name",
        field_type=ExportFieldType.STRING,
        column=SessionRow.domain_name,
    ),
    ExportFieldDef(
        key="access_key",
        name="Access Key",
        description="Owning access key",
        field_type=ExportFieldType.STRING,
        column=SessionRow.access_key,
    ),
    ExportFieldDef(
        key="status",
        name="Status",
        description="Session status",
        field_type=ExportFieldType.ENUM,
        column=SessionRow.status,
        formatter=lambda v: str(v) if v else "",
    ),
    ExportFieldDef(
        key="status_info",
        name="Status Info",
        description="Status details",
        field_type=ExportFieldType.STRING,
        column=SessionRow.status_info,
    ),
    ExportFieldDef(
        key="cluster_size",
        name="Cluster Size",
        description="Number of cluster nodes",
        field_type=ExportFieldType.INTEGER,
        column=SessionRow.cluster_size,
    ),
    ExportFieldDef(
        key="resource_used",
        name="Resources Used",
        description="Occupied resource slots",
        field_type=ExportFieldType.JSON,
        column=SessionRow.occupying_slots,
        formatter=lambda v: json.dumps(dict(v), default=str) if v else "",
    ),
    ExportFieldDef(
        key="resource_requested",
        name="Resources Requested",
        description="Requested resource slots",
        field_type=ExportFieldType.JSON,
        column=SessionRow.requested_slots,
        formatter=lambda v: json.dumps(dict(v), default=str) if v else "",
    ),
    ExportFieldDef(
        key="created_at",
        name="Created At",
        description="Session creation time",
        field_type=ExportFieldType.DATETIME,
        column=SessionRow.created_at,
        formatter=lambda v: v.isoformat() if v else "",
    ),
    ExportFieldDef(
        key="terminated_at",
        name="Terminated At",
        description="Session termination time",
        field_type=ExportFieldType.DATETIME,
        column=SessionRow.terminated_at,
        formatter=lambda v: v.isoformat() if v else "",
    ),
    # =========================================================================
    # Main Kernel Fields (N:1, no duplication)
    # =========================================================================
    ExportFieldDef(
        key="main_kernel_image",
        name="Main Kernel Image",
        description="Main kernel image canonical name",
        field_type=ExportFieldType.STRING,
        column=KernelRow.image,
        joins=frozenset({MAIN_KERNEL_JOIN}),
    ),
    ExportFieldDef(
        key="main_kernel_architecture",
        name="Main Kernel Architecture",
        description="Main kernel architecture",
        field_type=ExportFieldType.STRING,
        column=KernelRow.architecture,
        joins=frozenset({MAIN_KERNEL_JOIN}),
    ),
    ExportFieldDef(
        key="main_kernel_registry",
        name="Main Kernel Registry",
        description="Main kernel container registry",
        field_type=ExportFieldType.STRING,
        column=KernelRow.registry,
        joins=frozenset({MAIN_KERNEL_JOIN}),
    ),
    ExportFieldDef(
        key="main_kernel_tag",
        name="Main Kernel Tag",
        description="Main kernel image tag",
        field_type=ExportFieldType.STRING,
        column=KernelRow.tag,
        joins=frozenset({MAIN_KERNEL_JOIN}),
    ),
    # =========================================================================
    # Project Fields (N:1, no duplication)
    # =========================================================================
    ExportFieldDef(
        key="project_name",
        name="Project Name",
        description="Project (group) name",
        field_type=ExportFieldType.STRING,
        column=GroupRow.name,
        joins=frozenset({PROJECT_JOIN}),
    ),
    ExportFieldDef(
        key="project_description",
        name="Project Description",
        description="Project description",
        field_type=ExportFieldType.STRING,
        column=GroupRow.description,
        joins=frozenset({PROJECT_JOIN}),
    ),
    ExportFieldDef(
        key="project_resource_policy",
        name="Project Resource Policy",
        description="Project resource policy name",
        field_type=ExportFieldType.STRING,
        column=GroupRow.resource_policy,
        joins=frozenset({PROJECT_JOIN}),
    ),
    ExportFieldDef(
        key="project_is_active",
        name="Project Active",
        description="Project active status",
        field_type=ExportFieldType.BOOLEAN,
        column=GroupRow.is_active,
        joins=frozenset({PROJECT_JOIN}),
    ),
    ExportFieldDef(
        key="project_created_at",
        name="Project Created At",
        description="Project creation time",
        field_type=ExportFieldType.DATETIME,
        column=GroupRow.created_at,
        formatter=lambda v: v.isoformat() if v else "",
        joins=frozenset({PROJECT_JOIN}),
    ),
    # =========================================================================
    # Project Resource Policy Fields (N:1 through Project, no duplication)
    # =========================================================================
    ExportFieldDef(
        key="project_policy_max_vfolder_count",
        name="Project Policy Max VFolders",
        description="Maximum vfolder count for project",
        field_type=ExportFieldType.INTEGER,
        column=ProjectResourcePolicyRow.max_vfolder_count,
        joins=PROJECT_POLICY_JOINS,
    ),
    ExportFieldDef(
        key="project_policy_max_quota_scope_size",
        name="Project Policy Max Quota",
        description="Maximum quota scope size for project",
        field_type=ExportFieldType.INTEGER,
        column=ProjectResourcePolicyRow.max_quota_scope_size,
        joins=PROJECT_POLICY_JOINS,
    ),
    ExportFieldDef(
        key="project_policy_max_network_count",
        name="Project Policy Max Networks",
        description="Maximum network count for project",
        field_type=ExportFieldType.INTEGER,
        column=ProjectResourcePolicyRow.max_network_count,
        joins=PROJECT_POLICY_JOINS,
    ),
    # =========================================================================
    # User Fields (N:1, no duplication)
    # =========================================================================
    ExportFieldDef(
        key="user_email",
        name="User Email",
        description="User email address",
        field_type=ExportFieldType.STRING,
        column=UserRow.email,
        joins=frozenset({USER_JOIN}),
    ),
    ExportFieldDef(
        key="user_username",
        name="User Username",
        description="Username",
        field_type=ExportFieldType.STRING,
        column=UserRow.username,
        joins=frozenset({USER_JOIN}),
    ),
    ExportFieldDef(
        key="user_full_name",
        name="User Full Name",
        description="User full name",
        field_type=ExportFieldType.STRING,
        column=UserRow.full_name,
        joins=frozenset({USER_JOIN}),
    ),
    ExportFieldDef(
        key="user_role",
        name="User Role",
        description="User role (admin, user, etc.)",
        field_type=ExportFieldType.STRING,
        column=UserRow.role,
        formatter=lambda v: str(v) if v else "",
        joins=frozenset({USER_JOIN}),
    ),
    # =========================================================================
    # Resource Group Fields (N:1, no duplication)
    # =========================================================================
    ExportFieldDef(
        key="resource_group_name",
        name="Resource Group Name",
        description="Resource group name",
        field_type=ExportFieldType.STRING,
        column=ScalingGroupRow.name,
        joins=frozenset({SCALING_GROUP_JOIN}),
    ),
    ExportFieldDef(
        key="resource_group_description",
        name="Resource Group Description",
        description="Resource group description",
        field_type=ExportFieldType.STRING,
        column=ScalingGroupRow.description,
        joins=frozenset({SCALING_GROUP_JOIN}),
    ),
    ExportFieldDef(
        key="resource_group_is_active",
        name="Resource Group Active",
        description="Resource group active status",
        field_type=ExportFieldType.BOOLEAN,
        column=ScalingGroupRow.is_active,
        joins=frozenset({SCALING_GROUP_JOIN}),
    ),
    ExportFieldDef(
        key="resource_group_is_public",
        name="Resource Group Public",
        description="Resource group public status",
        field_type=ExportFieldType.BOOLEAN,
        column=ScalingGroupRow.is_public,
        joins=frozenset({SCALING_GROUP_JOIN}),
    ),
    ExportFieldDef(
        key="resource_group_scheduler",
        name="Resource Group Scheduler",
        description="Resource group scheduler type",
        field_type=ExportFieldType.STRING,
        column=ScalingGroupRow.scheduler,
        joins=frozenset({SCALING_GROUP_JOIN}),
    ),
    ExportFieldDef(
        key="resource_group_created_at",
        name="Resource Group Created At",
        description="Resource group creation time",
        field_type=ExportFieldType.DATETIME,
        column=ScalingGroupRow.created_at,
        formatter=lambda v: v.isoformat() if v else "",
        joins=frozenset({SCALING_GROUP_JOIN}),
    ),
    # =========================================================================
    # Kernel Fields (1:N, causes duplication)
    # =========================================================================
    ExportFieldDef(
        key="kernel_id",
        name="Kernel ID",
        description="Kernel UUID",
        field_type=ExportFieldType.UUID,
        column=KernelRow.id,
        joins=frozenset({KERNEL_JOIN}),
    ),
    ExportFieldDef(
        key="kernel_role",
        name="Kernel Role",
        description="Kernel cluster role",
        field_type=ExportFieldType.STRING,
        column=KernelRow.cluster_role,
        joins=frozenset({KERNEL_JOIN}),
    ),
    ExportFieldDef(
        key="kernel_status",
        name="Kernel Status",
        description="Kernel status",
        field_type=ExportFieldType.STRING,
        column=KernelRow.status,
        formatter=lambda v: str(v) if v else "",
        joins=frozenset({KERNEL_JOIN}),
    ),
    ExportFieldDef(
        key="kernel_image",
        name="Kernel Image",
        description="Kernel image name",
        field_type=ExportFieldType.STRING,
        column=KernelRow.image,
        joins=frozenset({KERNEL_JOIN}),
    ),
    ExportFieldDef(
        key="kernel_agent",
        name="Kernel Agent",
        description="Agent ID running this kernel",
        field_type=ExportFieldType.STRING,
        column=KernelRow.agent,
        joins=frozenset({KERNEL_JOIN}),
    ),
    ExportFieldDef(
        key="kernel_created_at",
        name="Kernel Created At",
        description="Kernel creation time",
        field_type=ExportFieldType.DATETIME,
        column=KernelRow.created_at,
        formatter=lambda v: v.isoformat() if v else "",
        joins=frozenset({KERNEL_JOIN}),
    ),
    ExportFieldDef(
        key="kernel_terminated_at",
        name="Kernel Terminated At",
        description="Kernel termination time",
        field_type=ExportFieldType.DATETIME,
        column=KernelRow.terminated_at,
        formatter=lambda v: v.isoformat() if v else "",
        joins=frozenset({KERNEL_JOIN}),
    ),
]


# Report definition
SESSION_REPORT = ReportDef(
    report_key="sessions",
    name="Sessions",
    description="Compute session export report",
    select_from=SessionRow.__table__,
    fields=SESSION_FIELDS,
)



from pathlib import Path
from pprint import pformat
from typing import Any, Self

from pydantic import BaseModel, Field

from ai.backend.common.configs.etcd import EtcdConfig
from ai.backend.common.configs.loader import (
    ConfigOverrider,
    EnvLoader,
    LoaderChain,
    TomlConfigLoader,
)
from ai.backend.common.configs.pyroscope import PyroscopeConfig
from ai.backend.logging.config import LoggingConfig
from ai.backend.logging.types import LogLevel
from ai.backend.manager.config.unified import (
    DatabaseConfig,
    DebugConfig,
    ManagerConfig,
)

from .constant import MANAGER_LOCAL_CFG_OVERRIDE_ENVS


# TODO: Remove useless config fields from this
class BootstrapConfig(BaseModel):
    db: DatabaseConfig = Field(
        default_factory=DatabaseConfig,  # type: ignore[arg-type]
        description="""
        Database configuration settings.
        Defines how the manager connects to its PostgreSQL database.
        Contains connection details, credentials, and pool settings.
        """,
    )
    etcd: EtcdConfig = Field(
        default_factory=EtcdConfig,  # type: ignore[arg-type]
        description="""
        Etcd configuration settings.
        Used for distributed coordination between manager instances.
        Contains connection details and authentication information.
        """,
    )
    manager: ManagerConfig = Field(
        default_factory=ManagerConfig,  # type: ignore[arg-type]
        description="""
        Core manager service configuration.
        Controls how the manager operates, communicates, and scales.
        Includes network settings, process management, and service parameters.
        """,
    )
    logging: LoggingConfig = Field(
        default_factory=LoggingConfig,  # type: ignore[arg-type]
        description="""
        Logging system configuration.
        Controls how logs are formatted, filtered, and stored.
        Detailed configuration is handled in ai.backend.logging.
        """,
    )
    pyroscope: PyroscopeConfig = Field(
        default_factory=PyroscopeConfig,  # type: ignore[arg-type]
        description="""
        Pyroscope profiling configuration.
        Controls integration with the Pyroscope performance profiling tool.
        Used for monitoring and analyzing application performance.
        """,
    )
    debug: DebugConfig = Field(
        default_factory=DebugConfig,  # type: ignore[arg-type]
        description="""
        Debugging options configuration.
        Controls various debugging features and tools.
        Should typically be disabled in production environments.
        """,
    )

    def __repr__(self) -> str:
        return pformat(self.model_dump())

    @classmethod
    async def load_from_file(cls, config_path: Path, log_level: LogLevel = LogLevel.NOTSET) -> Self:
        """
        Load configuration from a config file.

        All configurations are loaded by a single LoaderChain in config_provider_ctx.
        But BootstrapConfig must be loaded before config_provider_ctx is invoked since they are required for the manager boot process.
        """

        overrides: list[tuple[tuple[str, ...], Any]] = [
            (("debug", "enabled"), log_level == LogLevel.DEBUG),
        ]
        if log_level != LogLevel.NOTSET:
            overrides += [
                (("logging", "level"), log_level),
                (("logging", "pkg-ns", "ai.backend"), log_level),
            ]

        file_loader = TomlConfigLoader(config_path, "manager")
        env_loader = EnvLoader(MANAGER_LOCAL_CFG_OVERRIDE_ENVS)
        cfg_overrider = ConfigOverrider(overrides)
        cfg_loader = LoaderChain([
            file_loader,
            env_loader,
            cfg_overrider,
        ])
        raw_cfg = await cfg_loader.load()

        return cls.model_validate(raw_cfg, by_name=True)



"""DataLoader functions for scheduling history entities."""

from __future__ import annotations

import uuid
from collections.abc import Sequence

from ai.backend.manager.data.deployment.types import DeploymentHistoryData, RouteHistoryData
from ai.backend.manager.data.session.types import SessionSchedulingHistoryData
from ai.backend.manager.repositories.base import BatchQuerier, OffsetPagination
from ai.backend.manager.repositories.scheduling_history.options import (
    DeploymentHistoryConditions,
    RouteHistoryConditions,
    SessionSchedulingHistoryConditions,
)
from ai.backend.manager.services.scheduling_history.actions.search_deployment_history import (
    SearchDeploymentHistoryAction,
)
from ai.backend.manager.services.scheduling_history.actions.search_route_history import (
    SearchRouteHistoryAction,
)
from ai.backend.manager.services.scheduling_history.actions.search_session_history import (
    SearchSessionHistoryAction,
)
from ai.backend.manager.services.scheduling_history.processors import SchedulingHistoryProcessors


async def load_session_histories_by_ids(
    processor: SchedulingHistoryProcessors,
    history_ids: Sequence[uuid.UUID],
) -> list[SessionSchedulingHistoryData | None]:
    """Batch load session scheduling histories by their IDs."""
    if not history_ids:
        return []

    querier = BatchQuerier(
        pagination=OffsetPagination(limit=len(history_ids)),
        conditions=[SessionSchedulingHistoryConditions.by_ids(history_ids)],
    )

    action_result = await processor.search_session_history.wait_for_complete(
        SearchSessionHistoryAction(querier=querier)
    )

    history_map = {h.id: h for h in action_result.histories}
    return [history_map.get(history_id) for history_id in history_ids]


async def load_deployment_histories_by_ids(
    processor: SchedulingHistoryProcessors,
    history_ids: Sequence[uuid.UUID],
) -> list[DeploymentHistoryData | None]:
    """Batch load deployment histories by their IDs."""
    if not history_ids:
        return []

    querier = BatchQuerier(
        pagination=OffsetPagination(limit=len(history_ids)),
        conditions=[DeploymentHistoryConditions.by_ids(history_ids)],
    )

    action_result = await processor.search_deployment_history.wait_for_complete(
        SearchDeploymentHistoryAction(querier=querier)
    )

    history_map = {h.id: h for h in action_result.histories}
    return [history_map.get(history_id) for history_id in history_ids]


async def load_route_histories_by_ids(
    processor: SchedulingHistoryProcessors,
    history_ids: Sequence[uuid.UUID],
) -> list[RouteHistoryData | None]:
    """Batch load route histories by their IDs."""
    if not history_ids:
        return []

    querier = BatchQuerier(
        pagination=OffsetPagination(limit=len(history_ids)),
        conditions=[RouteHistoryConditions.by_ids(history_ids)],
    )

    action_result = await processor.search_route_history.wait_for_complete(
        SearchRouteHistoryAction(querier=querier)
    )

    history_map = {h.id: h for h in action_result.histories}
    return [history_map.get(history_id) for history_id in history_ids]



from __future__ import annotations

import logging
import re
import uuid
from collections.abc import Mapping, MutableMapping, Sequence
from dataclasses import dataclass
from typing import TYPE_CHECKING, Any, Self, cast
from urllib.parse import urlparse

import sqlalchemy as sa
import yarl
from sqlalchemy.exc import NoResultFound
from sqlalchemy.ext.asyncio import AsyncSession
from sqlalchemy.orm import Mapped, foreign, load_only, mapped_column, relationship

from ai.backend.common.container_registry import ContainerRegistryType
from ai.backend.common.exception import UnknownImageRegistry
from ai.backend.logging import BraceStyleAdapter
from ai.backend.manager.data.container_registry.types import ContainerRegistryData
from ai.backend.manager.errors.container_registry import (
    InvalidContainerRegistryProject,
    InvalidContainerRegistryURL,
)
from ai.backend.manager.models.base import (
    GUID,
    Base,
    StrEnumType,
)

if TYPE_CHECKING:
    from ai.backend.manager.models.association_container_registries_groups import (
        AssociationContainerRegistriesGroupsRow,
    )
    from ai.backend.manager.models.image import ImageRow

log = BraceStyleAdapter(logging.getLogger(__spec__.name))

__all__: Sequence[str] = (
    "ContainerRegistryRow",
    "ContainerRegistryValidator",
    "ContainerRegistryValidatorArgs",
)


@dataclass
class ContainerRegistryValidatorArgs:
    url: str
    type: ContainerRegistryType
    project: str | None


# TODO: Refactor this using inheritance
class ContainerRegistryValidator:
    """
    Validator for container registry configuration.
    """

    _url: str
    _type: ContainerRegistryType
    _project: str | None

    def __init__(self, args: ContainerRegistryValidatorArgs) -> None:
        self._url = args.url
        self._type = args.type
        self._project = args.project

    def _is_valid_url(self, url: str) -> bool:
        try:
            url = url.strip()
            if not url.startswith("http://") and not url.startswith("https://"):
                url = "http://" + url
            result = urlparse(url)
            return all([result.scheme, result.netloc])
        except Exception:
            return False

    def validate(self) -> None:
        """
        Validate container registry configuration.
        """
        # Validate URL format
        if not self._is_valid_url(self._url):
            raise InvalidContainerRegistryURL(f"Invalid URL format: {self._url}")

        # Validate project name for Harbor
        match self._type:
            case ContainerRegistryType.HARBOR | ContainerRegistryType.HARBOR2:
                if self._project is None:
                    raise InvalidContainerRegistryProject("Project name is required for Harbor.")
                if not (1 <= len(self._project) <= 255):
                    raise InvalidContainerRegistryProject("Invalid project name length.")
                pattern = re.compile(r"^[a-z0-9]+(?:[._-][a-z0-9]+)*$")
                if not pattern.match(self._project):
                    raise InvalidContainerRegistryProject("Invalid project name format.")
            case _:
                pass


def _get_image_join_condition() -> sa.ColumnElement[bool]:
    from ai.backend.manager.models.image import ImageRow

    return ContainerRegistryRow.id == foreign(ImageRow.registry_id)


def _get_association_join_condition() -> sa.ColumnElement[bool]:
    from ai.backend.manager.models.association_container_registries_groups import (
        AssociationContainerRegistriesGroupsRow,
    )

    return ContainerRegistryRow.id == foreign(AssociationContainerRegistriesGroupsRow.registry_id)


class ContainerRegistryRow(Base):  # type: ignore[misc]
    __tablename__ = "container_registries"

    id: Mapped[uuid.UUID] = mapped_column(
        "id", GUID, primary_key=True, server_default=sa.text("uuid_generate_v4()")
    )
    url: Mapped[str] = mapped_column("url", sa.String(length=512), index=True, nullable=False)
    registry_name: Mapped[str] = mapped_column(
        "registry_name", sa.String(length=255), index=True, nullable=False
    )
    type: Mapped[ContainerRegistryType] = mapped_column(
        "type",
        StrEnumType(ContainerRegistryType),
        default=ContainerRegistryType.DOCKER,
        server_default=ContainerRegistryType.DOCKER,
        nullable=False,
        index=True,
    )
    project: Mapped[str | None] = mapped_column(
        "project", sa.String(length=255), index=True, nullable=True
    )
    username: Mapped[str | None] = mapped_column("username", sa.String(length=255), nullable=True)
    password: Mapped[str | None] = mapped_column("password", sa.String, nullable=True)
    ssl_verify: Mapped[bool | None] = mapped_column(
        "ssl_verify", sa.Boolean, nullable=True, server_default=sa.text("true"), index=True
    )
    is_global: Mapped[bool | None] = mapped_column(
        "is_global", sa.Boolean, nullable=True, server_default=sa.text("true"), index=True
    )
    extra: Mapped[dict[str, Any] | None] = mapped_column(
        "extra", sa.JSON, nullable=True, default=None
    )

    image_rows: Mapped[list[ImageRow]] = relationship(
        "ImageRow",
        back_populates="registry_row",
        primaryjoin=_get_image_join_condition,
    )

    association_container_registries_groups_rows: Mapped[
        list[AssociationContainerRegistriesGroupsRow]
    ] = relationship(
        "AssociationContainerRegistriesGroupsRow",
        back_populates="container_registry_row",
        primaryjoin=_get_association_join_condition,
    )

    def __init__(
        self,
        id: uuid.UUID,
        url: str,
        registry_name: str,
        type: ContainerRegistryType,
        project: str | None = None,
        username: str | None = None,
        password: str | None = None,
        ssl_verify: bool | None = None,
        is_global: bool | None = None,
        extra: dict[str, Any] | None = None,
    ) -> None:
        self.id = id
        self.url = url
        self.registry_name = registry_name
        self.type = type
        self.project = project
        self.username = username
        self.password = password
        self.ssl_verify = ssl_verify
        self.is_global = is_global
        self.extra = extra

    @classmethod
    async def get(
        cls,
        session: AsyncSession,
        id: str | uuid.UUID,
    ) -> ContainerRegistryRow:
        query = sa.select(ContainerRegistryRow).where(ContainerRegistryRow.id == id)
        result = await session.execute(query)
        row = result.scalar()
        if row is None:
            raise NoResultFound
        return row

    @classmethod
    async def list_by_registry_name(
        cls,
        session: AsyncSession,
        registry_name: str,
    ) -> Sequence[ContainerRegistryRow]:
        query = sa.select(ContainerRegistryRow).where(
            ContainerRegistryRow.registry_name == registry_name
        )
        result = await session.execute(query)
        rows = result.scalars().all()
        if not rows:
            raise NoResultFound
        return rows

    @classmethod
    async def get_container_registry_info(
        cls, session: AsyncSession, registry_id: uuid.UUID
    ) -> tuple[yarl.URL, dict[str, Any]]:
        query_stmt = (
            sa.select(ContainerRegistryRow)
            .where(ContainerRegistryRow.id == registry_id)
            .options(
                load_only(
                    ContainerRegistryRow.url,
                    ContainerRegistryRow.username,
                    ContainerRegistryRow.password,
                )
            )
        )
        registry_row = cast(ContainerRegistryRow | None, await session.scalar(query_stmt))
        if registry_row is None:
            raise UnknownImageRegistry(registry_id)
        url = registry_row.url
        username = registry_row.username
        password = registry_row.password
        creds = {"username": username, "password": password}

        return yarl.URL(url), creds

    @classmethod
    async def get_known_container_registries(
        cls,
        session: AsyncSession,
    ) -> Mapping[str, Mapping[str, yarl.URL]]:
        query_stmt = sa.select(ContainerRegistryRow).options(
            load_only(
                ContainerRegistryRow.project,
                ContainerRegistryRow.registry_name,
                ContainerRegistryRow.url,
            )
        )
        registries = cast(list[ContainerRegistryRow], (await session.scalars(query_stmt)).all())
        result: MutableMapping[str, MutableMapping[str, yarl.URL]] = {}
        for registry_row in registries:
            project = registry_row.project
            if project is None:
                continue
            registry_name = registry_row.registry_name
            url = registry_row.url
            if project not in result:
                result[project] = {}
            result[project][registry_name] = yarl.URL(url)
        return result

    @classmethod
    def from_dataclass(cls, data: ContainerRegistryData) -> Self:
        instance = cls(
            id=data.id,
            url=data.url,
            registry_name=data.registry_name,
            type=data.type,
            project=data.project,
            username=data.username,
            password=data.password,
            ssl_verify=data.ssl_verify,
            is_global=data.is_global,
            extra=data.extra,
        )
        instance.id = data.id
        return instance

    def to_dataclass(self) -> ContainerRegistryData:
        return ContainerRegistryData(
            id=self.id,
            url=self.url,
            registry_name=self.registry_name,
            type=self.type,
            project=self.project,
            username=self.username,
            password=self.password,
            ssl_verify=self.ssl_verify,
            is_global=self.is_global,
            extra=self.extra,
        )



from __future__ import annotations

from ai.backend.common.exception import (
    BackendAIError,
    ErrorCode,
    ErrorDetail,
    ErrorDomain,
    ErrorOperation,
)


class RecorderError(BackendAIError):
    """Base exception for recorder errors."""

    error_type = "https://api.backend.ai/probs/recorder-error"
    error_title = "Recorder operation failed."

    def error_code(self) -> ErrorCode:
        return ErrorCode(
            domain=ErrorDomain.SESSION,
            operation=ErrorOperation.GENERIC,
            error_detail=ErrorDetail.INTERNAL_ERROR,
        )


class NestedPhaseError(RecorderError):
    """Raised when attempting to nest phases."""

    error_type = "https://api.backend.ai/probs/nested-phase-error"
    error_title = "Nested phases are not allowed."

    def __init__(self, new_phase: str, active_phase: str) -> None:
        self.new_phase = new_phase
        self.active_phase = active_phase
        super().__init__(
            extra_msg=(
                f"Phase '{new_phase}' cannot be started: phase '{active_phase}' is already active."
            )
        )

    def error_code(self) -> ErrorCode:
        return ErrorCode(
            domain=ErrorDomain.SESSION,
            operation=ErrorOperation.GENERIC,
            error_detail=ErrorDetail.BAD_REQUEST,
        )


class StepWithoutPhaseError(RecorderError):
    """Raised when attempting to record a step without an active phase."""

    error_type = "https://api.backend.ai/probs/step-without-phase-error"
    error_title = "Step requires an active phase."

    def __init__(self, step_name: str) -> None:
        self.step_name = step_name
        super().__init__(
            extra_msg=(
                f"Step '{step_name}' cannot be recorded: no active phase. "
                "Steps must be called within a phase context."
            )
        )

    def error_code(self) -> ErrorCode:
        return ErrorCode(
            domain=ErrorDomain.SESSION,
            operation=ErrorOperation.GENERIC,
            error_detail=ErrorDetail.BAD_REQUEST,
        )



# Models package for Backend.AI Manager
#
# IMPORTANT: Do NOT import symbols from this package directly.
# Use direct subpackage imports instead:
#   from ai.backend.manager.models.session import SessionRow
#   from ai.backend.manager.models.kernel import KernelRow
#
# For metadata with all tables registered:
#   from ai.backend.manager.models.metadata import metadata
